Package: fuzzystring
Title: Fast Fuzzy String Joins for Data Frames
Version: 0.0.6
Authors@R: 
    c(person(given = "Paul E.",
           family = "Santos Andrade",
           role = c("aut", "cre", "cph"),
           email = "paulefrens@gmail.com",
           comment = c(ORCID = "0000-0002-6635-0375")),
           person(given = "David",
             family = "Robinson",
             role = "ctb",
             email = "admiral.david@gmail.com",
             comment = "aut of fuzzyjoin"))
Description: Perform fuzzy joins on data frames using approximate string
    matching. Implements inner, left, right, full, semi, and anti joins with
    string distance metrics from the 'stringdist' package, including Optimal
    String Alignment, Levenshtein, Damerau-Levenshtein, Jaro-Winkler, q-gram,
    cosine, Jaccard, and Soundex. Uses a 'data.table' backend plus compiled
    'C++' result assembly to reduce overhead in large joins, while adaptive
    candidate planning avoids unnecessary distance evaluations in single-column
    string joins. Suitable for reconciling misspellings, inconsistent labels,
    and other near-match identifiers while optionally returning the computed
    distance for each match. Bibliographic references include the 'stringdist'
    package documentation (2014) <https://CRAN.R-project.org/package=stringdist>
    and Robinson, D. (2015) <https://github.com/dgrtwo/fuzzyjoin>.
License: MIT + file LICENSE
Depends: R (>= 4.1)
Imports: data.table, Rcpp, stringdist
LinkingTo: Rcpp
Suggests: dplyr, ggplot2, knitr, qdapDictionaries, readr, rmarkdown,
        rvest, stringr, testthat (>= 3.0.0), tibble, tidyr
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
URL: https://github.com/PaulESantos/fuzzystring,
        https://paulesantos.github.io/fuzzystring/
BugReports: https://github.com/PaulESantos/fuzzystring/issues
VignetteBuilder: knitr
Maintainer: Paul E. Santos Andrade <paulefrens@gmail.com>
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-07-24 01:38:59 UTC; PC
Author: Paul E. Santos Andrade [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-6635-0375>),
  David Robinson [ctb] (aut of fuzzyjoin)
Repository: CRAN
Date/Publication: 2026-07-24 06:40:02 UTC

# Generated by roxygen2: do not edit by hand

export(fuzzystring_anti_join)
export(fuzzystring_full_join)
export(fuzzystring_inner_join)
export(fuzzystring_join)
export(fuzzystring_left_join)
export(fuzzystring_right_join)
export(fuzzystring_semi_join)
importFrom(Rcpp,evalCpp)
importFrom(data.table,":=")
importFrom(data.table,.I)
importFrom(data.table,.N)
importFrom(data.table,CJ)
importFrom(data.table,as.data.table)
importFrom(data.table,copy)
importFrom(data.table,dcast)
importFrom(data.table,is.data.table)
importFrom(data.table,melt)
importFrom(data.table,rbindlist)
importFrom(data.table,setnames)
importFrom(data.table,setorder)
importFrom(stats,as.formula)
useDynLib(fuzzystring, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# fuzzystring <a href="https://paulesantos.github.io/fuzzystring/"><img src="man/figures/fuzzystring.png" align="right" width="200" style="max-width: 200px; height: auto;" /></a>

<!-- badges: start -->

[![CRAN status](https://www.r-pkg.org/badges/version/fuzzystring)](https://CRAN.R-project.org/package=fuzzystring)
[![R-CMD-check](https://github.com/PaulESantos/fuzzystring/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/PaulESantos/fuzzystring/actions/workflows/R-CMD-check.yaml)
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-blue.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![](https://cranlogs.r-pkg.org/badges/grand-total/fuzzystring?color=blue)](https://cran.r-project.org/package=fuzzystring)
[![](https://cranlogs.r-pkg.org/badges/last-week/fuzzystring?color=blue)](https://cran.r-project.org/package=fuzzystring)
<!-- badges: end -->

**fuzzystring** provides fast, flexible fuzzy string joins for `data.frame` and
`data.table` objects using approximate string matching. It combines
`stringdist`-based matching with a `data.table` backend and compiled C++ result
assembly to reduce overhead in large joins while preserving standard join
semantics.

## Why fuzzystring?

Real-world identifiers rarely line up exactly. `fuzzystring` is designed for
workloads such as:

- matching customer or company names with typos
- reconciling product catalogs with inconsistent labels
- linking survey responses to a controlled vocabulary
- joining reference tables to messy user input

The package includes:

- fuzzy `inner`, `left`, `right`, `full`, `semi`, and `anti` joins
- multiple `stringdist` methods, including OSA, Levenshtein,
  Damerau-Levenshtein, Jaro-Winkler, q-gram, cosine, jaccard, and soundex
- output that preserves the class of `x` (`data.table`, tibble, or base
  `data.frame`)
- optional distance columns for matched pairs
- case-insensitive matching
- adaptive candidate planning for single-column joins
- compiled C++ row expansion and result assembly across join modes

## Installation

``` r
# Install from CRAN
install.packages("fuzzystring")

# Development version from GitHub
# pak::pak("PaulESantos/fuzzystring")
# remotes::install_github("PaulESantos/fuzzystring")
```

## Quick start

``` r
library(fuzzystring)

x <- data.frame(
  name = c("Idea", "Premiom", "Very Good"),
  id = 1:3
)

y <- data.frame(
  approx_name = c("Ideal", "Premium", "VeryGood"),
  grp = c("A", "B", "C")
)

fuzzystring_inner_join(
  x, y,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)
```

## Join families

``` r
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), max_dist = 2)
fuzzystring_left_join(x, y, by = c(name = "approx_name"), max_dist = 2)
fuzzystring_right_join(x, y, by = c(name = "approx_name"), max_dist = 2)
fuzzystring_full_join(x, y, by = c(name = "approx_name"), max_dist = 2)
fuzzystring_semi_join(x, y, by = c(name = "approx_name"), max_dist = 2)
fuzzystring_anti_join(x, y, by = c(name = "approx_name"), max_dist = 2)
```

## Distance methods

``` r
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "osa")
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "dl")
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "jw")
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "soundex")
```

## Case-insensitive matching

``` r
fuzzystring_inner_join(
  x, y,
  by = c(name = "approx_name"),
  ignore_case = TRUE,
  max_dist = 1
)
```

## Included example data

The package ships with `misspellings`, a dataset of common misspellings adapted
from Wikipedia for examples and testing.

``` r
data(misspellings)
head(misspellings)
```

## Performance

`fuzzystring` keeps more of the join execution on a compiled path than the
original `fuzzyjoin` implementation. In practice, the package combines:

- `data.table` grouping and candidate planning
- adaptive blocking for single-column string joins
- compiled row expansion, row binding, and final assembly
- type-preserving handling of dates, datetimes, factors, and list-columns

The benchmark article summarizes a precomputed comparison against
`fuzzyjoin::stringdist_join()` using the same methods and sample sizes:

- Getting started: <https://paulesantos.github.io/fuzzystring/articles/getting_started.html>
- Benchmark article: <https://paulesantos.github.io/fuzzystring/articles/benchmark_fuzzyjoin_comparison.html>

## Multiple-column joins

`fuzzystring_join()` can match across more than one string column by applying
the same distance method and threshold to each mapped column.

``` r
x_multi <- data.frame(
  first = c("Jon", "Maira"),
  last = c("Smyth", "Gonzales")
)

y_multi <- data.frame(
  first_ref = c("John", "Maria"),
  last_ref = c("Smith", "Gonzalez"),
  id = 1:2
)

fuzzystring_inner_join(
  x_multi, y_multi,
  by = c(first = "first_ref", last = "last_ref"),
  method = "osa",
  max_dist = 1
)
```

## Related packages

- **[fuzzyjoin](https://github.com/dgrtwo/fuzzyjoin)**: original fuzzy join API
  that inspired this package
- **[stringdist](https://github.com/markvanderloo/stringdist)**: distance
  metrics
- **[data.table](https://rdatatable.gitlab.io/data.table/)**: high-performance
  tabular backend

## Credits

`fuzzystring` builds on ideas popularized by `fuzzyjoin`, while reinterpreting
the join pipeline around `data.table` and compiled C++ result assembly.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(fuzzystring)

## ----quick-start--------------------------------------------------------------
# Your messy data
x <- data.frame(
  name = c("Idea", "Premiom", "Very Good"), 
  id = 1:3
)

# Reference data
y <- data.frame(
  approx_name = c("Ideal", "Premium", "VeryGood"), 
  grp = c("A", "B", "C")
)

# Fuzzy join with max distance of 2 edits
fuzzystring_inner_join(
  x, y,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)

## ----join-datasets------------------------------------------------------------
x_join <- data.frame(
  name = c("Idea", "Premiom", "Very Good", "Gooood"),
  id = 1:4
)

y_join <- data.frame(
  approx_name = c("Ideal", "Premium", "VeryGood", "Good"),
  grp = c("A", "B", "C", "D")
)

## ----join-inner, eval = TRUE--------------------------------------------------
fuzzystring_inner_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)

## ----join-left, eval = TRUE---------------------------------------------------
fuzzystring_left_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)

## ----join-right, eval = TRUE--------------------------------------------------
fuzzystring_right_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)

## ----join-full, eval = TRUE---------------------------------------------------
fuzzystring_full_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)

## ----join-semi, eval = TRUE---------------------------------------------------
fuzzystring_semi_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2
)

## ----join-anti, eval = TRUE---------------------------------------------------
fuzzystring_anti_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2
)

## ----join-generic, eval = TRUE------------------------------------------------
fuzzystring_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  mode = "left",
  distance_col = "distance"
)

## ----distance-methods, eval = FALSE-------------------------------------------
# # Optimal String Alignment (default)
# fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "osa")
# 
# # Damerau-Levenshtein
# fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "dl")
# 
# # Jaro-Winkler (good for names)
# fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "jw")
# 
# # Soundex (phonetic matching)
# fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "soundex")

## ----ignore-case, eval = FALSE------------------------------------------------
# fuzzystring_inner_join(
#   x, y,
#   by = c(name = "approx_name"),
#   ignore_case = TRUE,
#   max_dist = 1
# )

## ----multi-column, eval = FALSE-----------------------------------------------
# x_multi <- data.frame(
#   first = c("Jon", "Maira"),
#   last = c("Smyth", "Gonzales")
# )
# 
# y_multi <- data.frame(
#   first_ref = c("John", "Maria"),
#   last_ref = c("Smith", "Gonzalez"),
#   customer_id = 1:2
# )
# 
# fuzzystring_inner_join(
#   x_multi, y_multi,
#   by = c(first = "first_ref", last = "last_ref"),
#   method = "osa",
#   max_dist = 1
# )

#' fuzzystring: Fast fuzzy string joins for data frames
#'
#' `fuzzystring` provides fuzzy inner, left, right, full, semi, and anti joins
#' for `data.frame` and `data.table` objects using approximate string matching.
#' It combines `stringdist` metrics with a `data.table` backend and compiled 'C++'
#' result assembly to reduce overhead in large joins while preserving familiar
#' join semantics.
#'
#' Main entry points are [fuzzystring_join()] and the convenience wrappers
#' [fuzzystring_inner_join()], [fuzzystring_left_join()],
#' [fuzzystring_right_join()], [fuzzystring_full_join()],
#' [fuzzystring_semi_join()], and [fuzzystring_anti_join()].
#'
#' The package also includes the example dataset [misspellings].
#'
#' @keywords package
#' @name fuzzystring
#' @docType package
#' @useDynLib fuzzystring, .registration = TRUE
#' @importFrom Rcpp evalCpp
"_PACKAGE"

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 12,
  fig.height = 8
)

## ----setup--------------------------------------------------------------------
library(dplyr)
library(tidyr)
library(ggplot2)

find_benchmark_csv <- function() {
  candidates <- c(
    file.path("inst", "extdata", "rbase_string_benchmark.csv"),
    file.path("..", "inst", "extdata", "rbase_string_benchmark.csv"),
    system.file("extdata", "rbase_string_benchmark.csv", package = "fuzzystring")
  )

  hit <- candidates[file.exists(candidates)][1]

  if (is.na(hit)) {
    stop(
      "Could not find the precomputed benchmark CSV. Expected it in ",
      paste(candidates, collapse = " or "),
      call. = FALSE
    )
  }

  hit
}

results_raw <- utils::read.csv(find_benchmark_csv(), stringsAsFactors = FALSE)

method_levels <- c(
  "osa", "lv", "dl", "hamming", "lcs",
  "qgram", "cosine", "jaccard", "jw", "soundex"
)

method_labels <- c(
  osa = "OSA",
  lv = "Levenshtein",
  dl = "Damerau-Levenshtein",
  hamming = "Hamming",
  lcs = "LCS",
  qgram = "Q-gram",
  cosine = "Cosine",
  jaccard = "Jaccard",
  jw = "Jaro-Winkler",
  soundex = "Soundex"
)

comparison_levels <- sprintf(
  "%.2f",
  sort(unique(results_raw$n_comps / 1e6))
)

summary_raw <- results_raw %>%
  group_by(expr, method, n_comps) %>%
  summarise(mean_time_ns = mean(time), .groups = "drop") %>%
  mutate(mean_time_ms = mean_time_ns / 1e6)

wide_summary <- summary_raw %>%
  select(expr, method, n_comps, mean_time_ms) %>%
  pivot_wider(names_from = expr, values_from = mean_time_ms) %>%
  mutate(
    runtime_ratio = fuzzyjoin / fuzzystring,
    method_label = factor(
      recode(method, !!!method_labels),
      levels = unname(method_labels[method_levels])
    ),
    comparisons_millions = n_comps / 1e6,
    comparisons_label = factor(
      sprintf("%.2f", comparisons_millions),
      levels = comparison_levels
    )
  ) %>%
  arrange(method_label, n_comps)

absolute_plot_data <- summary_raw %>%
  mutate(
    time_seconds = mean_time_ms / 1000,
    method_label = factor(
      recode(method, !!!method_labels),
      levels = unname(method_labels[method_levels])
    ),
    comparisons_label = factor(
      sprintf("%.2f", n_comps / 1e6),
      levels = comparison_levels
    ),
    implementation = factor(expr, levels = c("fuzzyjoin", "fuzzystring"))
  )

method_ranking <- wide_summary %>%
  group_by(method_label) %>%
  summarise(avg_runtime_ratio = mean(runtime_ratio), .groups = "drop") %>%
  arrange(desc(avg_runtime_ratio))

summary_table <- wide_summary %>%
  transmute(
    Method = as.character(method_label),
    `Candidate comparisons (M)` = sprintf("%.2f", comparisons_millions),
    `Mean time: fuzzyjoin (ms)` = round(fuzzyjoin, 2),
    `Mean time: fuzzystring (ms)` = round(fuzzystring, 2),
    `Runtime ratio (fuzzyjoin / fuzzystring)` = round(runtime_ratio, 2)
  )

ranking_table <- method_ranking %>%
  transmute(
    Method = as.character(method_label),
    `Average runtime ratio` = round(avg_runtime_ratio, 2)
  )

overall_ratio <- mean(wide_summary$runtime_ratio)

## ----benchmark-script, eval = FALSE-------------------------------------------
# library(microbenchmark)
# library(fuzzyjoin)
# library(fuzzystring)
# library(qdapDictionaries)
# library(tibble)
# 
# samp_sizes <- c(250, 500, 750)
# seed <- 2016
# 
# params <- list(
#   list(method = "osa",     mode = "inner", max_dist = 1,   q = 0),
#   list(method = "lv",      mode = "inner", max_dist = 1,   q = 0),
#   list(method = "dl",      mode = "inner", max_dist = 1,   q = 0),
#   list(method = "hamming", mode = "inner", max_dist = 1,   q = 0),
#   list(method = "lcs",     mode = "inner", max_dist = 1,   q = 0),
#   list(method = "qgram",   mode = "inner", max_dist = 2,   q = 2),
#   list(method = "cosine",  mode = "inner", max_dist = 0.5, q = 2),
#   list(method = "jaccard", mode = "inner", max_dist = 0.5, q = 2),
#   list(method = "jw",      mode = "inner", max_dist = 0.5, q = 0),
#   list(method = "soundex", mode = "inner", max_dist = 0.5, q = 0)
# )
# 
# args <- commandArgs(trailingOnly = TRUE)
# if (length(args) > 0) {
#   params <- Filter(function(p) p$method %in% args, params)
# }
# 
# data(misspellings)
# words <- as.data.frame(DICTIONARY)
# 
# results <- data.frame()
# 
# for (p in params) {
#   for (nsamp in samp_sizes) {
#     cat(sprintf("Running %s with %d samples\n", p$method, nsamp))
# 
#     set.seed(seed)
#     sub_misspellings <- misspellings[sample(seq_len(nrow(misspellings)), nsamp), ]
# 
#     bench <- microbenchmark(
#       fuzzyjoin = {
#         fuzzy_res <- stringdist_join(
#           sub_misspellings, words,
#           by         = c(misspelling = "word"),
#           method     = p$method,
#           mode       = p$mode,
#           max_dist   = p$max_dist,
#           q          = p$q
#         )
#       },
#       fuzzystring = {
#         fuzzystring_res <- fuzzystring_join(
#           sub_misspellings, words,
#           by       = c(misspelling = "word"),
#           method   = p$method,
#           mode     = p$mode,
#           max_dist = p$max_dist,
#           q        = p$q
#         )
#       },
#       times = 10
#     )
# 
#     # Verify that both implementations return the same result.
#     fuzzy_res <- stringdist_join(
#       sub_misspellings, words,
#       by       = c(misspelling = "word"),
#       method   = p$method,
#       mode     = p$mode,
#       max_dist = p$max_dist,
#       q        = p$q
#     )
# 
#     fuzzystring_res <- fuzzystring_join(
#       sub_misspellings, words,
#       by       = c(misspelling = "word"),
#       method   = p$method,
#       mode     = p$mode,
#       max_dist = p$max_dist,
#       q        = p$q
#     )
# 
#     if (!isTRUE(all.equal(fuzzy_res, fuzzystring_res)) && p$method != "soundex") {
#       message("Mismatch detected: fuzzyjoin vs fuzzystring for method: ", p$method)
#     }
# 
#     df <- as.data.frame(bench)
#     df$method <- p$method
#     df$n_comps <- nrow(sub_misspellings) * nrow(words)
#     df$os <- unname(Sys.info()["sysname"])
# 
#     results <- rbind(results, df)
#   }
# }
# 

## ----summary-table, echo=FALSE------------------------------------------------
knitr::kable(summary_table, align = "lrrrr")

## ----ranking-table, echo=FALSE------------------------------------------------
knitr::kable(ranking_table, align = "lr")

## ----absolute-runtime-plot----------------------------------------------------
ggplot(
  absolute_plot_data,
  aes(x = comparisons_label, y = time_seconds, fill = implementation)
) +
  geom_col(position = position_dodge(width = 0.75), width = 0.65) +
  facet_wrap(~ method_label, scales = "free_y") +
  labs(
    title = "Mean runtime by distance method",
    x = "Candidate comparisons (millions)",
    y = "Mean runtime (seconds)",
    fill = "Implementation"
  ) +
  scale_fill_manual(
    values = c("fuzzyjoin" = "#D95F02", "fuzzystring" = "#1B9E77"),
    labels = c("fuzzyjoin", "fuzzystring")
  ) +
  scale_y_continuous(labels = function(x) sprintf("%.1f", x)) +
  theme_minimal(base_size = 14) +
  theme(
    strip.text = element_text(face = "bold"),
    legend.position = "bottom",
    plot.title = element_text(face = "bold", hjust = 0.5),
    plot.subtitle = element_text(hjust = 0.5),
    axis.text.x = element_text(angle = 30, hjust = 1)
  )

## ----relative-runtime-plot----------------------------------------------------
ggplot(
  wide_summary,
  aes(x = comparisons_label, y = runtime_ratio)
) +
  geom_col(fill = "#D95F02", width = 0.55) +
  geom_hline(yintercept = 1, linetype = "dashed", color = "gray40") +
  facet_wrap(~ method_label, scales = "free_y") +
  labs(
    title = "Relative runtime against fuzzystring",
    subtitle = "Values above 1 mean fuzzystring is faster than fuzzyjoin",
    x = "Candidate comparisons (millions)",
    y = "Runtime ratio (fuzzyjoin / fuzzystring)"
  ) +
  scale_y_continuous(labels = function(x) sprintf("%.1f", x)) +
  theme_minimal(base_size = 14) +
  theme(
    strip.text = element_text(face = "bold"),
    legend.position = "none",
    plot.title = element_text(face = "bold", hjust = 0.5),
    plot.subtitle = element_text(hjust = 0.5),
    axis.text.x = element_text(angle = 30, hjust = 1)
  )

#' Fuzzy join backend using 'data.table' plus compiled 'C++' assembly
#'
#' Low-level engine used by \code{\link{fuzzystring_join}} and the compiled
#' fuzzy join helpers. It builds the match index with R 'data.table' and then
#' expands and assembles the result using compiled 'C++' binding code for speed.
#'
#' @param match_fun A function used to match values. It must return a logical
#'   vector (or a data.frame/data.table whose first column is logical) indicating
#'   which pairs match. For multi-column joins, you may pass a list of functions
#'   (one per column).
#' @param x A \code{data.frame} or \code{data.table}.
#' @param y A \code{data.frame} or \code{data.table}.
#' @param by Columns by which to join the two tables. See
#'   \code{\link{fuzzystring_join}}.
#' @param multi_by A character vector of column names used for multi-column
#'   matching when \code{multi_match_fun} is supplied.
#' @param multi_match_fun A function that receives matrices of unique values for
#'   \code{x} and \code{y} (rows correspond to unique combinations of \code{multi_by}).
#'   It must return a logical vector (or a data.frame/data.table whose first column
#'   is logical) indicating which rows match.
#' @param index_match_fun A function that receives the joined columns from
#'   \code{x} and \code{y} and returns a table with integer columns \code{x} and
#'   \code{y} (1-based row indices).
#' @param mode One of \code{"inner"}, \code{"left"}, \code{"right"}, \code{"full"},
#'   \code{"semi"}, or \code{"anti"}.
#' @param ... Additional arguments passed to the matching function(s).
#'
#' @return A joined table that preserves the container class of \code{x}:
#'   \code{data.table} inputs return \code{data.table}, tibble inputs return
#'   tibble, and plain \code{data.frame} inputs return plain
#'   \code{data.frame}. See \code{\link{fuzzystring_join}}.
#'
#' @details
#' This function works like \code{\link{fuzzystring_join}}, but replaces the
#' R-based output assembly with a compiled 'C++' implementation. This provides
#' better performance, especially for large joins, outer joins, and wide tables.
#' It is intended as a backend and does not compute distances itself; use
#' \code{\link{fuzzystring_join}} for string-distance based matching.
#'
#' The C++ implementation handles:
#' \itemize{
#'   \item Efficient subsetting by row indices
#'   \item Proper handling of NA values in outer joins
#'   \item Type-safe column operations for all common R types
#'   \item Preservation of factor levels and attributes
#'   \item Column name conflicts with .x/.y suffixes
#' }
#'
#' @importFrom stats as.formula
#' @importFrom data.table as.data.table is.data.table copy setorder setnames CJ rbindlist melt dcast .I .N :=
#' @keywords internal
fuzzystring_join_backend <- function(
  x,
  y,
  by = NULL,
  match_fun = NULL,
  multi_by = NULL,
  multi_match_fun = NULL,
  index_match_fun = NULL,
  mode = "inner",
  ...
) {
  mode <- match.arg(mode, c("inner", "left", "right", "full", "semi", "anti"))

  non_nulls <- (!is.null(multi_match_fun)) +
    (!is.null(match_fun)) +
    (!is.null(index_match_fun))
  if (sum(non_nulls) != 1) {
    stop(
      "Must give exactly one of match_fun, multi_match_fun, and index_match_fun",
      call. = FALSE
    )
  }

  x_is_dt <- data.table::is.data.table(x)
  x_is_tibble <- inherits(x, "tbl_df")

  restore_output_class <- function(out) {
    if (x_is_dt) {
      return(out)
    }

    if (x_is_tibble) {
      if (!requireNamespace("tibble", quietly = TRUE)) {
        stop(
          "Returning tibble output requires the 'tibble' package to be installed.",
          call. = FALSE
        )
      }
      return(tibble::as_tibble(out, .name_repair = "minimal"))
    }

    as.data.frame(out)
  }

  if (data.table::is.data.table(x)) {
    x_dt <- x
  } else {
    x_dt <- data.table::as.data.table(x)
  }

  if (data.table::is.data.table(y)) {
    y_dt <- y
  } else {
    y_dt <- data.table::as.data.table(y)
  }

  # --- helpers internos (same as dt_fuzzy_join) ------------------------------

  as_mapper_dt <- function(f) {
    if (inherits(f, "formula")) {
      stop(
        "Formula notation (~) not supported in this data.table version yet. Pass a function.",
        call. = FALSE
      )
    }
    if (!is.function(f)) {
      stop(
        "match_fun / multi_match_fun / index_match_fun must be a function.",
        call. = FALSE
      )
    }
    f
  }

  group_indices_1col <- function(dt, col) {
    # Group rows by values in a single column and return row indices for each group.
    # Used to identify which rows in x/y have the same value before matching.
    # Returns: data.table with columns (col_name, indices) where indices is a list of row numbers
    dt[, .(indices = list(.I)), by = c(col), verbose = FALSE]
  }

  group_indices_multicol <- function(dt, cols) {
    # Group rows by values in multiple columns and return row indices for each combination.
    # Used for multi-column fuzzy matching (e.g., match on both "first_name" and "last_name").
    # Returns: data.table with columns (cols..., indices) where indices is a list of row numbers
    dt[, .(indices = list(.I)), by = cols, verbose = FALSE]
  }

  expand_index_lists <- function(x_lists, y_lists) {
    # Compute Cartesian product of index groups for matched values.
    #
    # When two values match (e.g., "Alice" from x matches "Alice" from y),
    # all row pairs must be combined: if x has rows [1,3] and y has rows [2,5],
    # the output includes pairs (1,2), (1,5), (3,2), (3,5).
    #
    # Args:
    #   x_lists: list of integer vectors, each containing row indices for one value
    #   y_lists: corresponding list for y table (same length as x_lists)
    #
    # Returns: list with components:
    #   x: all row indices from x, repeated appropriately
    #   y: all row indices from y, repeated appropriately
    #   x_len, y_len: original group sizes (for validation/bookkeeping)

    x_len <- lengths(x_lists)
    y_len <- lengths(y_lists)

    sizes <- fst_checked_cartesian_sizes(x_len, y_len)
    rep_counts <- sizes$rep_counts
    total_size <- sizes$total_size

    if (total_size == 0L) {
      return(list(x = integer(0), y = integer(0), x_len = x_len, y_len = y_len))
    }

    x_rep <- integer(total_size)
    y_rep <- integer(total_size)

    pos <- 1L
    for (i in seq_along(x_lists)) {
      lx <- x_len[i]
      ly <- y_len[i]
      if (lx > 0L && ly > 0L) {
        n <- lx * ly
        end_pos <- pos + n - 1L
        # Repeat each x index ly times (expanding rows for all y in this group)
        x_rep[pos:end_pos] <- rep.int(x_lists[[i]], times = ly)
        # Repeat y indices each lx times (pairing each y with all x in this group)
        y_rep[pos:end_pos] <- rep(y_lists[[i]], each = lx)
        pos <- end_pos + 1L
      }
    }

    list(x = x_rep, y = y_rep, x_len = x_len, y_len = y_len)
  }

  replicate_extras <- function(extra_dt, w, x_len, y_len) {
    # Replicate extra columns (e.g., distance values) to match Cartesian product expansion.
    # When matching produces n Cartesian product rows from one group, the extra column value
    # (which is computed once per group) must be replicated n times.
    #
    # Args:
    #   extra_dt: data.table with extra columns (e.g., distance values), one row per match
    #   w: indices of matched groups (1-based, to index into extra_dt)
    #   x_len, y_len: group sizes used to compute rep_counts = x_len * y_len
    #
    # Returns: expanded extra_dt with rows replicated according to Cartesian product sizes
    if (is.null(extra_dt)) {
      return(NULL)
    }
    rep_counts <- x_len * y_len
    if (sum(rep_counts) == 0L) {
      return(extra_dt[0])
    }
    idx <- rep.int(w, times = rep_counts)
    extra_dt[idx]
  }

  # --- C++ binding function --------------------------------------------------
  # Uses the compiled C++ function instead of R implementation

  bind_by_rowid_cpp_wrapper <- function(
    x_dt2,
    y_dt2,
    matches_dt,
    overlap = character(0)
  ) {
    ret <- bind_by_rowid_cpp_matches(x_dt2, y_dt2, matches_dt, overlap)
    if (!data.table::is.data.table(ret)) {
      ret <- data.table::as.data.table(ret)
    }
    ret
  }

  # --- construir matches (same logic as dt_fuzzy_join) -----------------------

  matches <- NULL

  if (!is.null(match_fun)) {
    by2 <- fst_common_by(by, x_dt, y_dt)

    if (is.list(match_fun)) {
      mf_list <- lapply(match_fun, as_mapper_dt)
    } else {
      mf_list <- list(as_mapper_dt(match_fun))
    }

    if (length(mf_list) == 1L) {
      mf_list <- rep(mf_list, length(by2$x))
    }
    if (length(mf_list) != length(by2$x)) {
      stop(
        "Length of match_fun not equal to columns specified in 'by'.",
        call. = FALSE
      )
    }

    if (length(by2$x) > 1L) {
      # Multi-column progressive short-circuiting matching
      xcol <- by2$x[1L]
      ycol <- by2$y[1L]

      ix <- group_indices_1col(x_dt, xcol)
      iy <- group_indices_1col(y_dt, ycol)

      ux <- ix[[xcol]]
      uy <- iy[[ycol]]

      n_x <- length(ux)
      n_y <- length(uy)

      mf <- if (!is.null(names(mf_list))) mf_list[[xcol]] else mf_list[[1L]]

      m <- mf(rep(ux, times = n_y), rep(uy, each = n_x), ...)

      extra_dt <- NULL
      if (data.table::is.data.table(m) || is.data.frame(m)) {
        m_dt <- data.table::as.data.table(m)
        if (ncol(m_dt) > 1L) {
          extra_dt <- m_dt[, -1, with = FALSE]
        }
        m <- m_dt[[1]]
      }

      if (!is.logical(m)) {
        stop(
          "match_fun must return logical or a data.frame/data.table whose first column is logical.",
          call. = FALSE
        )
      }

      w <- which(m) - 1L
      if (length(w) == 0L) {
        matches <- data.table::data.table(x = integer(0), y = integer(0))
      } else {
        gx <- (w %% n_x) + 1L
        gy <- (w %/% n_x) + 1L

        x_lists <- ix$indices[gx]
        y_lists <- iy$indices[gy]

        expd <- expand_index_lists(x_lists, y_lists)
        matches <- data.table::data.table(x = expd$x, y = expd$y)

        if (!is.null(extra_dt)) {
          extra_rep <- replicate_extras(
            extra_dt,
            w + 1L,
            expd$x_len,
            expd$y_len
          )
          matches <- data.table::as.data.table(cbind(matches, extra_rep))
        }
      }

      if (nrow(matches) > 0L) {
        for (i in 2:length(by2$x)) {
          xcol <- by2$x[i]
          ycol <- by2$y[i]

          # Extract actual values for candidate row pairs
          v1 <- x_dt[[xcol]][matches$x]
          v2 <- y_dt[[ycol]][matches$y]

          mf <- if (!is.null(names(mf_list))) mf_list[[xcol]] else mf_list[[i]]
          m <- mf(v1, v2, ...)

          extra_dt <- NULL
          if (data.table::is.data.table(m) || is.data.frame(m)) {
            m_dt <- data.table::as.data.table(m)
            if (ncol(m_dt) > 1L) {
              extra_dt <- m_dt[, -1, with = FALSE]
            }
            m <- m_dt[[1]]
          }

          if (!is.logical(m)) {
            stop(
              "match_fun must return logical or a data.frame/data.table whose first column is logical.",
              call. = FALSE
            )
          }

          keep <- which(m)
          if (length(keep) == 0L) {
            matches <- data.table::data.table(x = integer(0), y = integer(0))
            break
          }

          matches <- matches[keep]

          if (!is.null(extra_dt)) {
            extra_rep <- extra_dt[keep]
            shared_extra <- intersect(
              setdiff(names(matches), c("x", "y")),
              names(extra_rep)
            )
            for (extra_col in shared_extra) {
              if (
                !is.numeric(matches[[extra_col]]) ||
                  !is.numeric(extra_rep[[extra_col]])
              ) {
                stop(
                  "Repeated match metadata columns must be numeric.",
                  call. = FALSE
                )
              }
              data.table::set(
                matches,
                j = extra_col,
                value = pmax(matches[[extra_col]], extra_rep[[extra_col]])
              )
            }
            new_extra <- setdiff(names(extra_rep), shared_extra)
            if (length(new_extra)) {
              matches <- data.table::as.data.table(cbind(
                matches,
                extra_rep[, new_extra, with = FALSE]
              ))
            }
          }
        }
      }
    } else {
      # Single-column match: remains exactly as original (fastest path for 1 col)
      xcol <- by2$x[1L]
      ycol <- by2$y[1L]

      ix <- group_indices_1col(x_dt, xcol)
      iy <- group_indices_1col(y_dt, ycol)

      ux <- ix[[xcol]]
      uy <- iy[[ycol]]

      n_x <- length(ux)
      n_y <- length(uy)

      mf <- if (!is.null(names(mf_list))) mf_list[[xcol]] else mf_list[[1L]]

      m <- mf(rep(ux, times = n_y), rep(uy, each = n_x), ...)

      extra_dt <- NULL
      if (data.table::is.data.table(m) || is.data.frame(m)) {
        m_dt <- data.table::as.data.table(m)
        if (ncol(m_dt) > 1L) {
          extra_dt <- m_dt[, -1, with = FALSE]
        }
        m <- m_dt[[1]]
      }

      if (!is.logical(m)) {
        stop(
          "match_fun must return logical or a data.frame/data.table whose first column is logical.",
          call. = FALSE
        )
      }

      w <- which(m) - 1L
      if (length(w) == 0L) {
        matches <- data.table::data.table(x = integer(0), y = integer(0))
      } else {
        gx <- (w %% n_x) + 1L
        gy <- (w %/% n_x) + 1L

        x_lists <- ix$indices[gx]
        y_lists <- iy$indices[gy]

        expd <- expand_index_lists(x_lists, y_lists)
        matches <- data.table::data.table(x = expd$x, y = expd$y)

        if (!is.null(extra_dt)) {
          extra_rep <- replicate_extras(
            extra_dt,
            w + 1L,
            expd$x_len,
            expd$y_len
          )
          matches <- data.table::as.data.table(cbind(matches, extra_rep))
        }
      }
    }
  } else if (!is.null(multi_match_fun)) {
    mmf <- as_mapper_dt(multi_match_fun)
    by2 <- fst_common_by(multi_by, x_dt, y_dt)

    ix <- group_indices_multicol(x_dt, by2$x)
    iy <- group_indices_multicol(y_dt, by2$y)

    ux <- as.matrix(ix[, by2$x, with = FALSE])
    uy <- as.matrix(iy[, by2$y, with = FALSE])

    n_x <- nrow(ux)
    n_y <- nrow(uy)

    if (n_x == 0L || n_y == 0L) {
      matches <- data.table::data.table(x = numeric(0), y = numeric(0))
    } else {
      grid <- data.table::CJ(ix = seq_len(n_x), iy = seq_len(n_y))
      ux_in <- ux[grid$ix, , drop = FALSE]
      uy_in <- uy[grid$iy, , drop = FALSE]

      m <- mmf(ux_in, uy_in)

      extra_dt <- NULL
      if (data.table::is.data.table(m) || is.data.frame(m)) {
        m_dt <- data.table::as.data.table(m)
        if (ncol(m_dt) > 1L) {
          extra_dt <- m_dt[, -1, with = FALSE]
        }
        m <- m_dt[[1]]
      }
      if (!is.logical(m)) {
        stop(
          "multi_match_fun must return logical or a data.frame/data.table whose first column is logical.",
          call. = FALSE
        )
      }

      if (sum(m) == 0L) {
        matches <- data.table::data.table(x = numeric(0), y = numeric(0))
      } else {
        keep <- which(m)
        x_lists <- ix$indices[grid$ix[keep]]
        y_lists <- iy$indices[grid$iy[keep]]

        expd <- expand_index_lists(x_lists, y_lists)
        matches <- data.table::data.table(x = expd$x, y = expd$y)

        if (!is.null(extra_dt)) {
          extra_rep <- replicate_extras(extra_dt, keep, expd$x_len, expd$y_len)
          matches <- data.table::as.data.table(cbind(matches, extra_rep))
        }
      }
    }
  } else {
    imf <- as_mapper_dt(index_match_fun)
    by2 <- fst_common_by(multi_by, x_dt, y_dt)

    d1 <- x_dt[, by2$x, with = FALSE]
    d2 <- y_dt[, by2$y, with = FALSE]
    matches <- imf(d1, d2)

    if (!data.table::is.data.table(matches)) {
      matches <- data.table::as.data.table(matches)
    }
    if (!all(c("x", "y") %in% names(matches))) {
      stop(
        "index_match_fun must return a table with columns named 'x' and 'y' (1-based row indices).",
        call. = FALSE
      )
    }
  }

  if (!data.table::is.data.table(matches)) {
    matches <- data.table::as.data.table(matches)
  }
  if ("i" %in% names(matches)) {
    matches[, i := NULL]
  }

  # --- modos semi / anti -----------------------------------------------------

  if (mode == "semi") {
    keep <- sort(unique(matches$x))
    keep <- keep[!is.na(keep)]
    res <- x_dt[keep]
    return(restore_output_class(res))
  }

  if (mode == "anti") {
    if (nrow(matches) == 0L) {
      return(restore_output_class(x_dt))
    }
    drop <- sort(unique(matches$x))
    drop <- drop[!is.na(drop)]
    res <- x_dt[-drop]
    return(restore_output_class(res))
  }

  # --- preparar renombres y completar índices según modo ---------------------

  data.table::setorder(matches, x, y)

  overlap <- intersect(names(x_dt), names(y_dt))
  if (length(overlap) > 0L) {
    if (data.table::is.data.table(x)) {
      x_dt <- data.table::copy(x_dt)
    }
    if (data.table::is.data.table(y)) {
      y_dt <- data.table::copy(y_dt)
    }
    data.table::setnames(x_dt, overlap, paste0(overlap, ".x"))
    data.table::setnames(y_dt, overlap, paste0(overlap, ".y"))
  }

  if (mode == "left") {
    allx <- data.table::data.table(x = seq_len(nrow(x_dt)))
    matches <- merge(allx, matches, by = "x", all.x = TRUE, sort = FALSE)
  } else if (mode == "right") {
    ally <- data.table::data.table(y = seq_len(nrow(y_dt)))
    matches <- merge(ally, matches, by = "y", all.x = TRUE, sort = FALSE)
  } else if (mode == "full") {
    has_x <- unique(matches[!is.na(x), x])
    has_y <- unique(matches[!is.na(y), y])

    miss_x <- setdiff(seq_len(nrow(x_dt)), has_x)
    miss_y <- setdiff(seq_len(nrow(y_dt)), has_y)

    if (length(miss_x)) {
      matches <- data.table::rbindlist(
        list(matches, data.table::data.table(x = miss_x, y = NA_integer_)),
        use.names = TRUE,
        fill = TRUE
      )
    }
    if (length(miss_y)) {
      matches <- data.table::rbindlist(
        list(matches, data.table::data.table(x = NA_integer_, y = miss_y)),
        use.names = TRUE,
        fill = TRUE
      )
    }
    data.table::setorder(matches, x, y)
  }

  # --- construir resultado final usando C++ ---------------------------------

  ret <- bind_by_rowid_cpp_wrapper(x_dt, y_dt, matches, overlap)
  ret <- fst_ensure_distance_col(ret, distance_col = NULL, mode = mode)

  restore_output_class(ret)
}
#

# This file is part of the standard setup for testthat.
# It is recommended that you do not modify it.
#
# Where should you do additional test configuration?
# Learn more about the roles of various files in:
# * https://r-pkgs.org/testing-design.html#sec-tests-files-overview
# * https://testthat.r-lib.org/articles/special-files.html

if (requireNamespace("testthat", quietly = TRUE)) {
  if (requireNamespace("data.table", quietly = TRUE)) {
    data.table::setDTthreads(1)
  }

  library(fuzzystring)
  testthat::test_check("fuzzystring")
}

#' A corpus of common misspellings, for examples and practice
#'
#' This is a \code{tbl_df} mapping misspellings of their words, compiled by
#' Wikipedia, where it is licensed under the CC-BY SA license. (Three words with
#' non-ASCII characters were filtered out). If you'd like to reproduce this
#' dataset from Wikipedia, see the example code below.
#'
#' @examples
#'
#' \donttest{
#' if (interactive()) {
#' library(rvest)
#' library(readr)
#' library(dplyr)
#' library(stringr)
#' library(tidyr)
#'
#' u <- "https://en.wikipedia.org/wiki/Wikipedia:Lists_of_common_misspellings/For_machines"
#' h <- read_html(u)
#'
#' misspellings <- h %>%
#'   html_nodes("pre") %>%
#'   html_text() %>%
#'   read_delim(col_names = c("misspelling", "correct"),
#'                     delim = ">",
#'                     skip = 1) %>%
#'   mutate(misspelling = str_sub(misspelling,
#'                                                1, -2)) |>
#'   separate_rows(correct, sep = ", ") |>
#'   filter(Encoding(correct) != "UTF-8")
#' }
#' }
#'
#' @source
#'   \url{https://en.wikipedia.org/wiki/Wikipedia:Lists_of_common_misspellings/For_machines}
"misspellings"

#' Null-coalescing operator (internal)
#'
#' Internal helper that returns \code{y} only when \code{x} is \code{NULL},
#' otherwise returns \code{x}. This is a small convenience used throughout the
#' package to express "use a default if NULL" semantics.
#'
#' @param x An object that may be \code{NULL}.
#' @param y A default value returned when \code{x} is \code{NULL}.
#'
#' @return \code{x} if it is not \code{NULL}; otherwise \code{y}.
#'
#' @keywords internal
#' @noRd
`%||%` <- function(x, y) if (is.null(x)) y else x

fst_checked_cartesian_sizes <- function(x_len, y_len) {
  rep_counts <- as.double(x_len) * as.double(y_len)
  total_size <- sum(rep_counts)

  if (!is.finite(total_size) || total_size > .Machine$integer.max) {
    stop(
      "The join result is too large to represent safely.",
      call. = FALSE
    )
  }

  list(
    rep_counts = as.integer(rep_counts),
    total_size = as.integer(total_size)
  )
}

fst_validate_join_inputs <- function(
  x,
  y,
  by,
  max_dist,
  ignore_case,
  distance_col
) {
  if (!is.data.frame(x) || !is.data.frame(y)) {
    stop("`x` and `y` must be data frames.", call. = FALSE)
  }
  if (
    !is.numeric(max_dist) ||
      length(max_dist) != 1L ||
      is.na(max_dist) ||
      !is.finite(max_dist) ||
      max_dist < 0
  ) {
    stop(
      "`max_dist` must be a single finite non-negative number.",
      call. = FALSE
    )
  }
  if (
    !is.logical(ignore_case) || length(ignore_case) != 1L || is.na(ignore_case)
  ) {
    stop("`ignore_case` must be `TRUE` or `FALSE`.", call. = FALSE)
  }
  if (!is.null(distance_col)) {
    valid_distance_col <- is.character(distance_col) &&
      length(distance_col) == 1L &&
      !is.na(distance_col) &&
      nzchar(distance_col)
    if (!valid_distance_col) {
      stop(
        "`distance_col` must be `NULL` or one non-empty string.",
        call. = FALSE
      )
    }
    unavailable <- c("x", "y", "i", names(x), names(y))
    if (distance_col %in% unavailable) {
      stop(
        "`distance_col` must not collide with input or internal columns.",
        call. = FALSE
      )
    }
  }

  by2 <- fst_common_by(by, x, y)
  valid_by <- is.list(by2) &&
    identical(sort(names(by2)), c("x", "y")) &&
    is.character(by2$x) &&
    is.character(by2$y) &&
    length(by2$x) > 0L &&
    length(by2$x) == length(by2$y) &&
    !anyNA(by2$x) &&
    !anyNA(by2$y) &&
    all(nzchar(by2$x)) &&
    all(nzchar(by2$y))
  if (!valid_by) {
    stop("`by` must specify at least one valid pair of columns.", call. = FALSE)
  }
  missing_x <- setdiff(by2$x, names(x))
  missing_y <- setdiff(by2$y, names(y))
  if (length(missing_x) || length(missing_y)) {
    missing <- character()
    if (length(missing_x)) {
      missing <- c(missing, paste0("x$", missing_x))
    }
    if (length(missing_y)) {
      missing <- c(missing, paste0("y$", missing_y))
    }
    stop(
      "Join columns not found: ",
      paste(missing, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  by2
}


#' Normalize join columns specification (data.table backend) (internal)
#'
#' Internal helper that standardizes a join specification into a list with
#' components \code{x} and \code{y}, containing the column names to use from
#' each input table.
#'
#' This mirrors the behavior of the \code{common_by()} helper used in
#' dplyr/tidyr-style joins, but is implemented using base R / data.table
#' conventions.
#'
#' @param by Join specification. May be:
#' \itemize{
#'   \item \code{NULL}: use the intersection of \code{names(x)} and \code{names(y)}.
#'   \item A character vector: column names used in both \code{x} and \code{y}.
#'   \item A named character vector: names are columns in \code{x}, values are columns in \code{y}.
#'   \item A list with components \code{x} and \code{y}: returned as-is.
#' }
#' @param x A \code{data.frame} or \code{data.table}.
#' @param y A \code{data.frame} or \code{data.table}.
#'
#' @details
#' If \code{by} is \code{NULL}, the function uses the intersection of column
#' names in \code{x} and \code{y}. If there are no common variables, it throws
#' an error and asks the caller to provide \code{by}.
#'
#' If \code{by} is a partially named character vector, unnamed entries (i.e.
#' \code{""} names) are assumed to refer to identical column names in both
#' tables.
#'
#' The function prints a message when joining by common variables (i.e. when
#' \code{by = NULL}), matching dplyr's behavior.
#'
#' @return A list with:
#' \itemize{
#'   \item \code{x}: character vector of column names in \code{x}
#'   \item \code{y}: character vector of column names in \code{y}
#' }
#'
#' @keywords internal
#' @noRd
fst_common_by <- function(by = NULL, x, y) {
  if (is.list(by)) {
    return(by)
  }

  if (!is.null(by)) {
    x_names <- names(by) %||% by
    y_names <- unname(by)

    # If x is partially named, assume blanks refer to identical names in y
    x_names[x_names == ""] <- y_names[x_names == ""]

    return(list(x = x_names, y = y_names))
  }

  common <- intersect(names(x), names(y))

  if (length(common) == 0) {
    stop("No common variables. Please specify `by` param.", call. = FALSE)
  }

  message("Joining by: ", paste(common, collapse = ", "))

  list(x = common, y = common)
}


#' Ensure distance column exists (data.table backend) (internal)
#'
#' Internal helper used by joins that optionally return a distance/similarity
#' metric (e.g. \code{stringdist_join()}). When \code{distance_col} is requested
#' and the join mode is not \code{"semi"} or \code{"anti"}, this function ensures
#' the output contains a column named \code{distance_col}.
#'
#' For \code{data.table} inputs, the column is added by reference via
#' \code{data.table::set()} for efficiency.
#'
#' @param ret A join result. Usually a \code{data.table}, but may be a
#'   \code{data.frame}.
#' @param distance_col A string naming the distance column to ensure, or
#'   \code{NULL} to do nothing.
#' @param mode Join mode. One of \code{"inner"}, \code{"left"}, \code{"right"},
#'   \code{"full"}, \code{"semi"}, or \code{"anti"}.
#'
#' @details
#' If \code{ret} already contains \code{distance_col}, it is returned unchanged.
#' Otherwise:
#' \itemize{
#'   \item If \code{nrow(ret) == 0}, an empty numeric vector is used.
#'   \item Else the column is filled with \code{NA}.
#' }
#'
#' No column is added for \code{mode = "semi"} or \code{mode = "anti"}, since
#' those modes return subsets of \code{x} only.
#'
#' @return \code{ret}, potentially modified to include \code{distance_col}.
#'
#' @keywords internal
#' @noRd
fst_ensure_distance_col <- function(ret, distance_col, mode) {
  if (!(mode %in% c("semi", "anti")) && !is.null(distance_col)) {
    if (is.null(ret[[distance_col]])) {
      val <- if (nrow(ret) == 0) numeric(0) else rep(NA_real_, nrow(ret))

      if (data.table::is.data.table(ret)) {
        # Asegurar que tiene espacio pre-alocado para nuevas columnas
        data.table::setalloccol(ret)
        data.table::set(ret, j = distance_col, value = val)
      } else {
        ret[[distance_col]] <- val
      }
    }
  }
  ret
}


#' Drop row names (internal)
#'
#' Internal helper that removes row names from a \code{data.frame}-like object.
#' This mirrors the behavior of the original implementation used to avoid
#' propagating row names through join results.
#'
#' For \code{data.table}, row names are not normally used, but some operations
#' may attach a \code{"rownames"} attribute. This helper normalizes the result
#' by resetting that attribute to the internal "row names" representation.
#'
#' @param x A \code{data.frame} or \code{data.table}.
#'
#' @return \code{x} with row names removed/reset.
#'
#' @keywords internal
#' @noRd
fst_unrowwname <- function(x) {
  if (data.table::is.data.table(x)) {
    data.table::setattr(x, "rownames", .set_row_names(nrow(x)))
  } else {
    rownames(x) <- NULL
  }
  x
}

fst_expand_match_indices <- function(x_lists, y_lists) {
  x_len <- lengths(x_lists)
  y_len <- lengths(y_lists)
  sizes <- fst_checked_cartesian_sizes(x_len, y_len)
  rep_counts <- sizes$rep_counts
  total_size <- sizes$total_size

  if (total_size == 0L) {
    return(list(x = integer(0), y = integer(0), rep_counts = integer(0)))
  }

  x_rep <- integer(total_size)
  y_rep <- integer(total_size)

  pos <- 1L
  for (i in seq_along(x_lists)) {
    lx <- x_len[i]
    ly <- y_len[i]
    if (lx == 0L || ly == 0L) {
      next
    }

    n <- lx * ly
    end_pos <- pos + n - 1L
    x_rep[pos:end_pos] <- rep.int(x_lists[[i]], times = ly)
    y_rep[pos:end_pos] <- rep(y_lists[[i]], each = lx)
    pos <- end_pos + 1L
  }

  list(x = x_rep, y = y_rep, rep_counts = rep_counts)
}

fst_chunk_size_y <- function(n_x, target_pairs = 250000L) {
  n_x <- max(1L, as.integer(n_x))
  max(1L, as.integer(target_pairs %/% n_x))
}

fst_is_length_pruned_method <- function(method) {
  method %in% c("osa", "lv", "dl", "hamming")
}

fst_can_use_exact_join <- function(method, max_dist, extra_args) {
  identical(max_dist, 0) &&
    method %in% c("osa", "lv", "dl", "hamming", "lcs", "jw") &&
    length(extra_args) == 0L
}

fst_exact_match_indices <- function(d1, d2, ignore_case, distance_col) {
  key_names <- paste0(".fst_key_", seq_along(d1))
  normalize <- function(data) {
    columns <- lapply(data, as.character)
    if (ignore_case) {
      columns <- lapply(columns, tolower)
    }
    result <- data.table::as.data.table(columns)
    data.table::setnames(result, key_names)
    result
  }

  x_keys <- normalize(d1)
  y_keys <- normalize(d2)
  data.table::set(x_keys, j = "x", value = seq_len(nrow(x_keys)))
  data.table::set(y_keys, j = "y", value = seq_len(nrow(y_keys)))
  x_keys <- x_keys[stats::complete.cases(x_keys)]
  y_keys <- y_keys[stats::complete.cases(y_keys)]

  if (nrow(x_keys) == 0L || nrow(y_keys) == 0L) {
    return(data.table::data.table(x = integer(), y = integer()))
  }

  matches <- merge(
    x_keys,
    y_keys,
    by = key_names,
    sort = FALSE,
    allow.cartesian = TRUE
  )
  matches <- matches[, c("x", "y"), with = FALSE]
  if (!is.null(distance_col)) {
    data.table::set(
      matches,
      j = distance_col,
      value = rep.int(0, nrow(matches))
    )
  }
  matches
}

fst_choose_target_pairs <- function(
  method,
  n_rows_x,
  n_rows_y,
  n_unique_x,
  n_unique_y,
  base_target_pairs = 250000L
) {
  if (fst_is_length_pruned_method(method)) {
    return(as.integer(base_target_pairs))
  }

  avg_dup_x <- n_rows_x / max(1L, n_unique_x)
  avg_dup_y <- n_rows_y / max(1L, n_unique_y)
  max_avg_dup <- max(avg_dup_x, avg_dup_y)

  if (max_avg_dup <= 1.10) {
    return(3000000L)
  }
  if (max_avg_dup <= 1.50) {
    return(2000000L)
  }
  if (max_avg_dup <= 2.00) {
    return(1000000L)
  }

  as.integer(base_target_pairs)
}

fst_use_dense_unique_block <- function(
  method,
  n_rows_x,
  n_rows_y,
  n_unique_x,
  n_unique_y,
  max_pairs = 12000000L
) {
  if (fst_is_length_pruned_method(method)) {
    return(FALSE)
  }

  if (n_unique_x == 0L || n_unique_y == 0L) {
    return(FALSE)
  }

  avg_dup_x <- n_rows_x / n_unique_x
  avg_dup_y <- n_rows_y / n_unique_y
  pair_count <- as.double(n_unique_x) * as.double(n_unique_y)

  pair_count <= max_pairs && avg_dup_x <= 2 && avg_dup_y <= 2
}

fst_stringdist_eval_block <- function(
  x_vals,
  y_vals,
  x_rows,
  y_rows,
  method,
  max_dist,
  distance_col,
  extra_args = list()
) {
  n_x <- length(x_vals)
  n_y <- length(y_vals)
  if (n_x == 0L || n_y == 0L) {
    return(data.table::data.table(x = integer(0), y = integer(0)))
  }

  x_rep <- rep.int(x_vals, times = n_y)
  y_rep <- rep(y_vals, each = n_x)
  dists <- do.call(
    stringdist::stringdist,
    c(list(a = x_rep, b = y_rep, method = method), extra_args)
  )

  keep <- (!is.na(dists)) & (dists <= max_dist)
  if (!any(keep)) {
    return(data.table::data.table(x = integer(0), y = integer(0)))
  }

  w <- which(keep) - 1L
  gx <- (w %% n_x) + 1L
  gy <- (w %/% n_x) + 1L

  x_lists <- x_rows[gx]
  y_lists <- y_rows[gy]
  expanded <- fst_expand_match_indices(x_lists, y_lists)

  out <- data.table::data.table(x = expanded$x, y = expanded$y)
  if (!is.null(distance_col)) {
    out[[distance_col]] <- rep.int(dists[keep], times = expanded$rep_counts)
  }

  out
}

fst_stringdist_single_col_matches <- function(
  v1,
  v2,
  max_dist,
  method,
  ignore_case = FALSE,
  distance_col = NULL,
  extra_args = list(),
  target_pairs = 250000L
) {
  v1 <- as.character(v1)
  v2 <- as.character(v2)

  if (isTRUE(ignore_case)) {
    v1 <- tolower(v1)
    v2 <- tolower(v2)
  }

  x_groups <- data.table::data.table(value = v1, row = seq_along(v1))
  y_groups <- data.table::data.table(value = v2, row = seq_along(v2))

  x_groups <- x_groups[!is.na(value), .(rows = list(row)), by = value]
  y_groups <- y_groups[!is.na(value), .(rows = list(row)), by = value]

  if (nrow(x_groups) == 0L || nrow(y_groups) == 0L) {
    return(data.table::data.table(x = integer(0), y = integer(0)))
  }

  x_groups[, grp := .I]
  y_groups[, grp := .I]

  use_length_pruning <- fst_is_length_pruned_method(method)
  target_pairs <- fst_choose_target_pairs(
    method = method,
    n_rows_x = length(v1),
    n_rows_y = length(v2),
    n_unique_x = nrow(x_groups),
    n_unique_y = nrow(y_groups),
    base_target_pairs = target_pairs
  )

  if (use_length_pruning) {
    x_groups[, nchars := nchar(value, type = "chars")]
    y_groups[, nchars := nchar(value, type = "chars")]
  }

  if (
    fst_use_dense_unique_block(
      method = method,
      n_rows_x = length(v1),
      n_rows_y = length(v2),
      n_unique_x = nrow(x_groups),
      n_unique_y = nrow(y_groups)
    )
  ) {
    return(fst_stringdist_eval_block(
      x_vals = x_groups$value,
      y_vals = y_groups$value,
      x_rows = x_groups$rows,
      y_rows = y_groups$rows,
      method = method,
      max_dist = max_dist,
      distance_col = distance_col,
      extra_args = extra_args
    ))
  }

  parts <- vector("list", 0L)

  if (use_length_pruning) {
    x_by_len <- split(x_groups$grp, x_groups$nchars)
    y_by_len <- split(y_groups$grp, y_groups$nchars)

    for (len_name in names(x_by_len)) {
      x_ids <- x_by_len[[len_name]]
      if (!length(x_ids)) {
        next
      }

      len_val <- as.integer(len_name)
      if (method == "hamming") {
        y_len_candidates <- as.character(len_val)
      } else {
        lower_len <- max(0L, as.integer(ceiling(len_val - max_dist)))
        upper_len <- as.integer(floor(len_val + max_dist))
        y_len_candidates <- as.character(seq.int(lower_len, upper_len))
      }
      y_len_candidates <- intersect(y_len_candidates, names(y_by_len))
      if (!length(y_len_candidates)) {
        next
      }

      for (y_len_name in y_len_candidates) {
        y_ids_all <- y_by_len[[y_len_name]]
        if (!length(y_ids_all)) {
          next
        }

        y_block_size <- fst_chunk_size_y(
          length(x_ids),
          target_pairs = target_pairs
        )
        y_splits <- split(
          y_ids_all,
          ceiling(seq_along(y_ids_all) / y_block_size)
        )

        for (y_ids in y_splits) {
          parts[[length(parts) + 1L]] <- fst_stringdist_eval_block(
            x_vals = x_groups$value[x_ids],
            y_vals = y_groups$value[y_ids],
            x_rows = x_groups$rows[x_ids],
            y_rows = y_groups$rows[y_ids],
            method = method,
            max_dist = max_dist,
            distance_col = distance_col,
            extra_args = extra_args
          )
        }
      }
    }
  } else {
    x_ids <- x_groups$grp
    y_ids_all <- y_groups$grp
    y_block_size <- fst_chunk_size_y(length(x_ids), target_pairs = target_pairs)
    y_splits <- split(y_ids_all, ceiling(seq_along(y_ids_all) / y_block_size))

    for (y_ids in y_splits) {
      parts[[length(parts) + 1L]] <- fst_stringdist_eval_block(
        x_vals = x_groups$value[x_ids],
        y_vals = y_groups$value[y_ids],
        x_rows = x_groups$rows[x_ids],
        y_rows = y_groups$rows[y_ids],
        method = method,
        max_dist = max_dist,
        distance_col = distance_col,
        extra_args = extra_args
      )
    }
  }

  if (!length(parts)) {
    return(data.table::data.table(x = integer(0), y = integer(0)))
  }

  data.table::rbindlist(parts, use.names = TRUE, fill = TRUE)
}

#' Join two tables based on fuzzy string matching
#'
#' Uses \code{stringdist::stringdist()} to compute distances and a
#' \code{data.table}-orchestrated backend with compiled 'C++' assembly to produce
#' the final result. This is the main user-facing entry point for fuzzy joins on
#' strings.
#'
#' @param x A \code{data.frame} or \code{data.table}.
#' @param y A \code{data.frame} or \code{data.table}.
#' @param by Columns by which to join the two tables. You can supply a character
#'   vector of common names (e.g. \code{c("name")} ), or a named vector mapping
#'   \code{x} to \code{y} (e.g. \code{c(name = "approx_name")}).
#' @param max_dist Maximum distance to use for joining. Smaller values are stricter.
#' @param ignore_case Logical; if \code{TRUE}, comparisons are case-insensitive.
#' @param method Method for computing string distance, see
#'   \code{?stringdist::stringdist} and the \code{stringdist} package vignettes.
#' @param distance_col If not \code{NULL}, adds a column with this name containing
#'   the computed distance for each matched pair (or \code{NA} for unmatched rows
#'   in outer joins).
#' @param mode One of \code{"inner"}, \code{"left"}, \code{"right"}, \code{"full"},
#'   \code{"semi"}, or \code{"anti"}.
#' @param ... Additional arguments passed to \code{\link[stringdist]{stringdist}}.
#'
#' @details
#' If \code{method = "soundex"}, \code{max_dist} is automatically set to 0.5,
#' since Soundex distance is 0 (match) or 1 (no match).
#'
#' When \code{by} maps multiple columns, the same \code{method},
#' \code{max_dist}, and any additional \code{stringdist} arguments are applied
#' independently to each mapped column, and a row pair is kept only when all
#' mapped columns satisfy the distance threshold.
#' When `distance_col` is requested for a multi-column join, it contains the
#' maximum distance across the mapped columns for each matched row pair.
#'
#' For single-column joins, fuzzystring uses adaptive candidate planning before
#' calling \code{stringdist::stringdist()}. For Levenshtein-like methods
#' (\code{"osa"}, \code{"lv"}, \code{"dl"}), a fast prefilter is applied: if
#' \code{abs(nchar(v1) - nchar(v2)) > max_dist}, the pair cannot match, so
#' distance is not computed for that pair. For low-duplication workloads, the
#' planner can also evaluate larger dense blocks of unique values to reduce
#' orchestration overhead while preserving the same matching semantics.
#' When `max_dist = 0` and no additional metric arguments are supplied,
#' methods whose zero distance implies string equality use an exact keyed join
#' instead of evaluating the full candidate distance set.
#'
#' @return A joined table that preserves the container class of \code{x}:
#'   \code{data.table} inputs return \code{data.table}, tibble inputs return
#'   tibble, and plain \code{data.frame} inputs return plain
#'   \code{data.frame}. See \code{\link{fuzzystring_join_backend}} for details
#'   on output structure.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("ggplot2", quietly = TRUE)) {
#'   d <- data.table::data.table(approximate_name = c("Idea", "Premiom"))
#'   # Match diamonds$cut to d$approximate_name
#'   res <- fuzzystring_inner_join(ggplot2::diamonds, d,
#'     by = c(cut = "approximate_name"),
#'     max_dist = 1
#'   )
#'   head(res)
#' }
#' }
#'
#' @export
fuzzystring_join <- function(
  x,
  y,
  by = NULL,
  max_dist = 2,
  method = c(
    "osa",
    "lv",
    "dl",
    "hamming",
    "lcs",
    "qgram",
    "cosine",
    "jaccard",
    "jw",
    "soundex"
  ),
  mode = "inner",
  ignore_case = FALSE,
  distance_col = NULL,
  ...
) {
  method <- match.arg(method)

  if (method == "soundex") {
    max_dist <- 0.5
  }

  by2 <- fst_validate_join_inputs(
    x,
    y,
    by,
    max_dist,
    ignore_case,
    distance_col
  )
  dot_args <- list(...)

  if (fst_can_use_exact_join(method, max_dist, dot_args)) {
    index_match_fun <- function(d1, d2) {
      fst_exact_match_indices(
        d1,
        d2,
        ignore_case = ignore_case,
        distance_col = distance_col
      )
    }
    out <- fuzzystring_join_backend(
      x,
      y,
      multi_by = by2,
      mode = mode,
      index_match_fun = index_match_fun
    )
  } else if (length(by2$x) == 1L) {
    index_match_fun <- function(d1, d2) {
      fst_stringdist_single_col_matches(
        d1[[1]],
        d2[[1]],
        max_dist = max_dist,
        method = method,
        ignore_case = ignore_case,
        distance_col = distance_col,
        extra_args = dot_args
      )
    }

    out <- fuzzystring_join_backend(
      x,
      y,
      multi_by = by2,
      mode = mode,
      index_match_fun = index_match_fun
    )
  } else {
    match_fun <- function(v1, v2) {
      v1 <- as.character(v1)
      v2 <- as.character(v2)

      if (isTRUE(ignore_case)) {
        v1 <- tolower(v1)
        v2 <- tolower(v2)
      }

      na_pair <- is.na(v1) | is.na(v2)

      if (method %in% c("osa", "lv", "dl")) {
        len1 <- ifelse(is.na(v1), NA_integer_, nchar(v1, type = "chars"))
        len2 <- ifelse(is.na(v2), NA_integer_, nchar(v2, type = "chars"))

        length_diff <- abs(len1 - len2)
        include <- (!na_pair) & (length_diff <= max_dist)

        dists <- rep(NA_real_, length(v1))
        if (any(include)) {
          dists[include] <- do.call(
            stringdist::stringdist,
            c(list(a = v1[include], b = v2[include], method = method), dot_args)
          )
        }
      } else {
        dists <- rep(NA_real_, length(v1))
        ok <- !na_pair
        if (any(ok)) {
          dists[ok] <- do.call(
            stringdist::stringdist,
            c(list(a = v1[ok], b = v2[ok], method = method), dot_args)
          )
        }
      }

      inc <- (!is.na(dists)) & (dists <= max_dist)

      ret <- data.table::data.table(include = inc)
      if (!is.null(distance_col)) {
        data.table::set(ret, j = distance_col, value = dists)
      }
      ret
    }

    out <- fuzzystring_join_backend(
      x,
      y,
      by = by2,
      mode = mode,
      match_fun = match_fun
    )
  }

  fst_ensure_distance_col(out, distance_col, mode)
}

#' @rdname fuzzystring_join
#' @export
fuzzystring_inner_join <- function(x, y, by = NULL, distance_col = NULL, ...) {
  fuzzystring_join(
    x,
    y,
    by = by,
    mode = "inner",
    distance_col = distance_col,
    ...
  )
}

#' @rdname fuzzystring_join
#' @export
fuzzystring_left_join <- function(x, y, by = NULL, distance_col = NULL, ...) {
  fuzzystring_join(
    x,
    y,
    by = by,
    mode = "left",
    distance_col = distance_col,
    ...
  )
}

#' @rdname fuzzystring_join
#' @export
fuzzystring_right_join <- function(x, y, by = NULL, distance_col = NULL, ...) {
  fuzzystring_join(
    x,
    y,
    by = by,
    mode = "right",
    distance_col = distance_col,
    ...
  )
}

#' @rdname fuzzystring_join
#' @export
fuzzystring_full_join <- function(x, y, by = NULL, distance_col = NULL, ...) {
  fuzzystring_join(
    x,
    y,
    by = by,
    mode = "full",
    distance_col = distance_col,
    ...
  )
}

#' @rdname fuzzystring_join
#' @export
fuzzystring_semi_join <- function(x, y, by = NULL, distance_col = NULL, ...) {
  fuzzystring_join(
    x,
    y,
    by = by,
    mode = "semi",
    distance_col = distance_col,
    ...
  )
}

#' @rdname fuzzystring_join
#' @export
fuzzystring_anti_join <- function(x, y, by = NULL, distance_col = NULL, ...) {
  fuzzystring_join(
    x,
    y,
    by = by,
    mode = "anti",
    distance_col = distance_col,
    ...
  )
}

test_that("exact route preserves duplicate and missing-value semantics", {
  x <- data.frame(name = c("Alpha", "beta", "Alpha", NA_character_))
  y <- data.frame(approx = c("Alpha", "gamma", "Alpha", NA_character_))

  for (method in c("osa", "lv", "dl", "hamming", "lcs", "jw")) {
    inner <- fuzzystring_inner_join(
      x,
      y,
      by = c(name = "approx"),
      method = method,
      max_dist = 0,
      distance_col = "distance"
    )

    expect_equal(nrow(inner), 4L, info = method)
    expect_equal(inner$distance, rep(0, 4L), info = method)
  }

  expect_equal(
    nrow(fuzzystring_left_join(x, y, by = c(name = "approx"), max_dist = 0)),
    6L
  )
  expect_equal(
    nrow(fuzzystring_right_join(x, y, by = c(name = "approx"), max_dist = 0)),
    6L
  )
  expect_equal(
    nrow(fuzzystring_full_join(x, y, by = c(name = "approx"), max_dist = 0)),
    8L
  )
  expect_equal(
    fuzzystring_semi_join(x, y, by = c(name = "approx"), max_dist = 0)$name,
    c("Alpha", "Alpha")
  )
  expect_equal(
    fuzzystring_anti_join(x, y, by = c(name = "approx"), max_dist = 0)$name,
    c("beta", NA_character_)
  )
})

test_that("exact route supports case folding and multiple columns", {
  x <- data.frame(
    first = c("Alice", "Bob", "ALICE"),
    last = c("Smith", "Jones", "Smith")
  )
  y <- data.frame(
    given = c("alice", "bob"),
    family = c("smith", "stone")
  )

  result <- fuzzystring_inner_join(
    x,
    y,
    by = c(first = "given", last = "family"),
    max_dist = 0,
    ignore_case = TRUE,
    distance_col = "distance"
  )

  expect_equal(result$first, c("Alice", "ALICE"))
  expect_equal(result$given, c("alice", "alice"))
  expect_equal(result$distance, c(0, 0))
})

test_that("exact route remains conservative when metric semantics may differ", {
  expect_identical(fst_can_use_exact_join("osa", 0, list()), TRUE)
  expect_identical(
    fst_can_use_exact_join("osa", 0, list(weight = rep(1, 4))),
    FALSE
  )
  expect_identical(fst_can_use_exact_join("qgram", 0, list()), FALSE)
  expect_identical(fst_can_use_exact_join("soundex", 0, list()), FALSE)
})

test_that("multi-column joins report the maximum component distance", {
  x <- data.frame(
    first = c("John", "Bob"),
    last = c("Smith", "Jones")
  )
  y <- data.frame(
    fname = c("Jon", "Bob"),
    lname = c("Smit", "Jones")
  )

  result <- fuzzystring_inner_join(
    x,
    y,
    by = c(first = "fname", last = "lname"),
    max_dist = 1,
    distance_col = "distance"
  )

  expect_named(result, c("first", "last", "fname", "lname", "distance"))
  expect_equal(result$distance, c(1, 0))
})

test_that("distance column cannot corrupt input or internal columns", {
  x <- data.frame(value = "foo")
  y <- data.frame(approx = "foo")

  expect_snapshot(
    error = TRUE,
    fuzzystring_inner_join(
      x,
      y,
      by = c(value = "approx"),
      distance_col = "x"
    )
  )
  expect_snapshot(
    error = TRUE,
    fuzzystring_inner_join(
      x,
      y,
      by = c(value = "approx"),
      distance_col = "value"
    )
  )
})

test_that("public arguments are validated before join planning", {
  x <- data.frame(value = "foo")
  y <- data.frame(approx = "foo")

  expect_snapshot(
    error = TRUE,
    fuzzystring_join(x, y, by = c(value = "approx"), max_dist = NA_real_)
  )
  expect_snapshot(
    error = TRUE,
    fuzzystring_join(x, y, by = c(value = "approx"), max_dist = Inf)
  )
  expect_snapshot(
    error = TRUE,
    fuzzystring_join(x, y, by = c(value = "approx"), ignore_case = "yes")
  )
  expect_snapshot(
    error = TRUE,
    fuzzystring_join(x, y, by = character())
  )
  expect_snapshot(
    error = TRUE,
    fuzzystring_join(x, y, by = c(missing = "approx"))
  )
})

test_that("Cartesian expansion rejects unsupported result sizes", {
  expect_snapshot(
    error = TRUE,
    fst_checked_cartesian_sizes(.Machine$integer.max, 2L)
  )
})

test_that("optimized single-column matching agrees with direct distances", {
  x_values <- c("alpha", "alfa", "beta", NA_character_, "alpha")
  y_values <- c("alpah", "beta", "delta", NA_character_)

  for (method in c("osa", "lv", "dl", "hamming", "lcs", "jw")) {
    actual <- fst_stringdist_single_col_matches(
      x_values,
      y_values,
      max_dist = 2,
      method = method,
      distance_col = "distance"
    )
    grid <- expand.grid(
      x = seq_along(x_values),
      y = seq_along(y_values)
    )
    grid$distance <- stringdist::stringdist(
      x_values[grid$x],
      y_values[grid$y],
      method = method
    )
    expected <- grid[!is.na(grid$distance) & grid$distance <= 2, ]
    attr(expected, "out.attrs") <- NULL

    actual <- actual[order(actual$x, actual$y), ]
    expected <- expected[order(expected$x, expected$y), ]
    rownames(expected) <- NULL
    expect_equal(as.data.frame(actual), expected, info = method)
  }
})

test_that("stringdist_join / variants (data.table backend)", {
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("stringdist")
  skip_if_not_installed("data.table")

  # Use ggplot2::diamonds (available when ggplot2 is installed)
  diamonds <- ggplot2::diamonds

  # ------------------------------------------------------------------
  # Setup helpers
  # ------------------------------------------------------------------
  d <- data.table::data.table(
    cut2 = c("Idea", "Premiums", "Premiom", "VeryGood", "VeryGood", "Faiir"),
    type = 1:6
  )

  d2 <- d[1:3]
  included <- c("Ideal", "Premium")
  notin <- c("Fair", "Good", "Very Good")

  d3 <- data.table::rbindlist(
    list(d2, data.table::data.table(cut2 = "NewType", type = 4L)),
    use.names = TRUE,
    fill = TRUE
  )

  # ------------------------------------------------------------------
  # INNER join (multiples + distance column)
  # ------------------------------------------------------------------
  j <- fuzzystring_inner_join(
    diamonds,
    d,
    by = c(cut = "cut2"),
    distance_col = "distance"
  )

  # count(cut, cut2) + arrange(cut)
  result <- data.table::as.data.table(j)[, .N, by = .(cut, cut2)]
  data.table::setorder(result, cut)

  expect_equal(
    as.character(result$cut),
    c("Fair", "Very Good", "Premium", "Premium", "Ideal")
  )
  expect_equal(
    result$cut2,
    c("Faiir", "VeryGood", "Premiums", "Premiom", "Idea")
  )

  # Premium and Very Good each match two entries in d (Premiom/Premiums, VeryGood twice)
  expect_equal(sum(j$cut == "Premium"), sum(diamonds$cut == "Premium") * 2L)
  expect_equal(sum(j$cut == "Very Good"), sum(diamonds$cut == "Very Good") * 2L)
  expect_equal(sum(j$cut2 == "Premiom"), sum(diamonds$cut == "Premium"))

  # With default max_dist=2, here all accepted should be distance 1
  expect_true(all(j$distance == 1))

  vg <- data.table::as.data.table(j)[cut == "Very Good", .N, by = type]
  data.table::setorder(vg, type)

  expect_equal(vg$type, c(4L, 5L))
  expect_equal(vg$N, rep.int(sum(diamonds$cut == "Very Good"), 2L))

  expect_true(all(j$type[j$cut2 == "Faiir"] == 6L))

  # ------------------------------------------------------------------
  # LEFT join
  # ------------------------------------------------------------------
  result_l <- fuzzystring_left_join(diamonds, d2, by = c(cut = "cut2"))

  expect_true(all(is.na(result_l$cut2[result_l$cut %in% notin])))
  expect_equal(sum(result_l$cut %in% notin), sum(diamonds$cut %in% notin))

  expect_equal(
    sum(result_l$cut2 == "Premiom", na.rm = TRUE),
    sum(diamonds$cut == "Premium")
  )
  expect_equal(
    sum(result_l$cut2 == "Premiom", na.rm = TRUE),
    sum(result_l$cut2 == "Premiums", na.rm = TRUE)
  )

  # ------------------------------------------------------------------
  # RIGHT join
  # ------------------------------------------------------------------
  result_r <- fuzzystring_right_join(diamonds, d3, by = c(cut = "cut2"))

  expect_equal(sum(result_r$cut2 == "NewType"), 1L)
  expect_equal(sum(is.na(result_r$cut)), 1L)
  expect_true(all(is.na(result_r$cut[result_r$cut2 == "NewType"])))

  expect_equal(
    sum(result_r$cut2 == "Premiom", na.rm = TRUE),
    sum(diamonds$cut == "Premium")
  )
  expect_equal(
    sum(result_r$cut2 == "Premiom", na.rm = TRUE),
    sum(result_r$cut2 == "Premiums", na.rm = TRUE)
  )

  # ------------------------------------------------------------------
  # FULL join
  # ------------------------------------------------------------------
  result_f <- fuzzystring_full_join(diamonds, d3, by = c(cut = "cut2"))

  expect_equal(sum(result_f$cut2 == "NewType", na.rm = TRUE), 1L)
  expect_equal(sum(is.na(result_f$cut)), 1L)
  expect_true(all(is.na(result_f$cut[result_f$cut2 == "NewType"])))

  expect_true(all(is.na(result_f$cut2[result_f$cut %in% notin])))
  expect_equal(sum(result_f$cut %in% notin), sum(diamonds$cut %in% notin))

  expect_equal(
    sum(result_f$cut2 == "Premiom", na.rm = TRUE),
    sum(diamonds$cut == "Premium")
  )
  expect_equal(
    sum(result_f$cut2 == "Premiom", na.rm = TRUE),
    sum(result_f$cut2 == "Premiums", na.rm = TRUE)
  )

  # ------------------------------------------------------------------
  # SEMI join
  # ------------------------------------------------------------------
  result_s <- fuzzystring_semi_join(diamonds, d2, by = c(cut = "cut2"))

  expect_equal(sort(as.character(unique(result_s$cut))), included)
  expect_equal(nrow(result_s), sum(result_s$cut %in% included))
  expect_false("cut2" %in% names(result_s))

  # ------------------------------------------------------------------
  # ANTI join
  # ------------------------------------------------------------------
  result_a <- fuzzystring_anti_join(diamonds, d2, by = c(cut = "cut2"))

  expect_equal(sort(as.character(unique(result_a$cut))), notin)
  expect_equal(nrow(result_a), sum(result_a$cut %in% notin))

  # ------------------------------------------------------------------
  # No matches cases (including overlapping column names)
  # ------------------------------------------------------------------
  d_nomatch <- data.table::data.table(
    cut2 = c(
      "Ideolll",
      "Premiumsss",
      "Premiomzzz",
      "VeryVeryGood",
      "VeryVeryGood",
      "FaiirsFair"
    ),
    type = 1:6
  )

  j1 <- fuzzystring_inner_join(diamonds, d_nomatch, by = c(cut = "cut2"))
  expect_equal(nrow(j1), 0L)
  expect_true(all(c("carat", "cut", "cut2", "type") %in% names(j1)))

  d_nomatch2 <- data.table::copy(d_nomatch)
  data.table::setnames(d_nomatch2, "cut2", "cut")
  j1_5 <- fuzzystring_inner_join(diamonds, d_nomatch2, by = c(cut = "cut"))
  expect_equal(nrow(j1_5), 0L)
  expect_true(all(c("carat", "cut.x", "cut.y", "type") %in% names(j1_5)))

  j2 <- fuzzystring_left_join(diamonds, d_nomatch, by = c(cut = "cut2"))
  expect_equal(nrow(j2), nrow(diamonds))
  expect_true(all(is.na(j2$cut2)))

  j3 <- fuzzystring_right_join(diamonds, d_nomatch, by = c(cut = "cut2"))
  expect_equal(nrow(j3), nrow(d_nomatch))
  expect_true(all(is.na(j3$carat)))
  expect_true(all(is.na(j3$cut)))

  j4 <- fuzzystring_full_join(diamonds, d_nomatch, by = c(cut = "cut2"))
  expect_equal(nrow(j4), nrow(diamonds) + nrow(d_nomatch))
  expect_true(all(is.na(j4$cut) | is.na(j4$cut2)))
  expect_true(all(is.na(j4$carat) | is.na(j4$type)))

  j5 <- fuzzystring_semi_join(diamonds, d_nomatch, by = c(cut = "cut2"))
  expect_equal(nrow(j5), 0L)
  expect_true("cut" %in% names(diamonds))
  expect_true("carat" %in% names(diamonds))
  expect_false("cut2" %in% names(diamonds))
  expect_false("type" %in% names(diamonds))

  j6 <- fuzzystring_anti_join(diamonds, d_nomatch, by = c(cut = "cut2"))
  expect_equal(nrow(j6), nrow(diamonds))
  expect_true("cut" %in% names(diamonds))
  expect_false("cut2" %in% names(diamonds))

  # ------------------------------------------------------------------
  # ignore_case
  # ------------------------------------------------------------------
  d_lowercase <- data.table::copy(d)
  d_lowercase[, cut2 := tolower(cut2)]

  j_ic_0 <- fuzzystring_inner_join(
    diamonds,
    d_lowercase,
    by = c(cut = "cut2"),
    distance_col = "distance",
    max_dist = 1
  )
  expect_equal(nrow(j_ic_0), 0L)

  j_ic <- fuzzystring_inner_join(
    diamonds,
    d_lowercase,
    by = c(cut = "cut2"),
    distance_col = "distance",
    ignore_case = TRUE,
    max_dist = 1
  )
  expect_gt(nrow(j_ic), 0L)
  expect_equal(sum(j_ic$cut == "Premium"), sum(diamonds$cut == "Premium") * 2L)
  expect_true(all(j_ic$distance[j_ic$cut2 == "idea"] == 1))

  # ------------------------------------------------------------------
  # soundex
  # ------------------------------------------------------------------
  j_sdx <- fuzzystring_inner_join(
    diamonds,
    d,
    by = c(cut = "cut2"),
    distance_col = "distance",
    method = "soundex"
  )
  expect_gt(nrow(j_sdx), 0L)
  expect_equal(sum(j_sdx$cut == "Premium"), sum(diamonds$cut == "Premium") * 2L)

  # ------------------------------------------------------------------
  # Renaming overlapping columns (.x / .y)
  # ------------------------------------------------------------------
  d_overlap <- data.table::data.table(
    cut = c("Idea", "Premiums", "Premiom", "VeryGood", "VeryGood", "Faiir"),
    price = 1:6
  )

  j_ov <- fuzzystring_inner_join(diamonds, d_overlap, by = "cut")

  expect_true("cut.x" %in% names(j_ov))
  expect_true("price.x" %in% names(j_ov))
  expect_true("cut.y" %in% names(j_ov))
  expect_true("price.y" %in% names(j_ov))
  expect_true(all(j_ov$cut.y %in% d_overlap$cut))
  expect_true(all(j_ov$price.y %in% d_overlap$price))

  # ------------------------------------------------------------------
  # Return type: data.frame in -> data.frame out
  # ------------------------------------------------------------------
  result_df <- fuzzystring_inner_join(
    as.data.frame(diamonds),
    d,
    by = c(cut = "cut2")
  )
  expect_s3_class(result_df, "data.frame")
  expect_false(inherits(result_df, "tbl_df"))

  result_df2 <- fuzzystring_inner_join(
    as.data.frame(diamonds),
    as.data.frame(d),
    by = c(cut = "cut2")
  )
  expect_s3_class(result_df2, "data.frame")
  expect_false(inherits(result_df2, "tbl_df"))

  # ------------------------------------------------------------------
  # One-column data.frames (regression)
  # ------------------------------------------------------------------
  onecol <- data.frame(
    cut2 = c("Idea", "Premiums", "Premiom", "VeryGood", "VeryGood", "Faiir")
  )
  diamonds_df <- as.data.frame(diamonds)
  res_onecol <- fuzzystring_inner_join(
    diamonds_df,
    onecol,
    by = c(cut = "cut2")
  )
  expect_s3_class(res_onecol, "data.frame")
  expect_true("cut2" %in% names(res_onecol))
  expect_gt(nrow(res_onecol), 0L)

  # ------------------------------------------------------------------
  # No common variables should error
  # ------------------------------------------------------------------
  expect_error(fuzzystring_inner_join(diamonds, d), "No common variables")

  # ------------------------------------------------------------------
  # Distance column existence when no overlaps (outer joins vs semi/anti)
  # ------------------------------------------------------------------
  a <- data.table::data.table(x = c("apple", "banana"))
  b <- data.table::data.table(y = c("orange", "mango"))

  res_left <- fuzzystring_left_join(
    a,
    b,
    by = c(x = "y"),
    max_dist = 1,
    distance_col = "distance"
  )
  expect_equal(names(res_left), c("x", "y", "distance"))
  expect_equal(nrow(res_left), 2L)
  expect_true(all(is.na(res_left$y)))
  expect_true(all(is.na(res_left$distance)))

  res_inner <- fuzzystring_inner_join(
    a,
    b,
    by = c(x = "y"),
    max_dist = 1,
    distance_col = "distance"
  )
  expect_equal(names(res_inner), c("x", "y", "distance"))
  expect_equal(nrow(res_inner), 0L)

  res_semi <- fuzzystring_semi_join(
    a,
    b,
    by = c(x = "y"),
    max_dist = 1,
    distance_col = "distance"
  )
  expect_equal(names(res_semi), "x")
  expect_equal(nrow(res_semi), 0L)

  res_anti <- fuzzystring_anti_join(
    a,
    b,
    by = c(x = "y"),
    max_dist = 1,
    distance_col = "distance"
  )
  expect_equal(as.data.frame(a), as.data.frame(res_anti))

  # ------------------------------------------------------------------
  # Overlapping names in outer joins keep .x / .y suffixes
  # ------------------------------------------------------------------
  x_overlap_outer <- data.table::data.table(
    key_value = c("alpha", "bravo"),
    shared = 1:2,
    when = as.Date(c("2024-01-01", "2024-01-02"))
  )
  data.table::setnames(x_overlap_outer, "key_value", "key")
  y_overlap_outer <- data.table::data.table(
    key_value = c("alphx", "charlie"),
    shared = 10:11,
    label = c("matched", "unmatched")
  )
  data.table::setnames(y_overlap_outer, "key_value", "key")

  outer_overlap <- fuzzystring_full_join(
    x_overlap_outer,
    y_overlap_outer,
    by = "key",
    max_dist = 1,
    distance_col = "distance"
  )

  expect_true(all(
    c(
      "key.x",
      "shared.x",
      "key.y",
      "shared.y",
      "when",
      "label",
      "distance"
    ) %in%
      names(outer_overlap)
  ))
  expect_s3_class(outer_overlap$when, "Date")
  expect_equal(sum(is.na(outer_overlap$key.x)), 1L)
  expect_equal(sum(is.na(outer_overlap$key.y)), 1L)

  # ------------------------------------------------------------------
  # Typed columns are preserved in compiled binding paths
  # ------------------------------------------------------------------
  x_typed <- data.table::data.table(
    name = c("alpha", "bravo"),
    when = as.Date(c("2024-01-01", "2024-01-02")),
    stamp = as.POSIXct(
      c("2024-01-01 10:00:00", "2024-01-02 11:30:00"),
      tz = "UTC"
    ),
    grp = factor(c("a", "b")),
    payload = I(list(list(id = 1L), list(id = 2L)))
  )
  y_typed <- data.table::data.table(
    approx_name = c("alphx", "charlie"),
    note = c("hit", "miss"),
    meta = I(list(list(code = 10L), list(code = 20L)))
  )

  typed_left <- fuzzystring_left_join(
    x_typed,
    y_typed,
    by = c(name = "approx_name"),
    max_dist = 1,
    distance_col = "distance"
  )

  expect_s3_class(typed_left$when, "Date")
  expect_s3_class(typed_left$stamp, "POSIXct")
  expect_equal(attr(typed_left$stamp, "tzone"), "UTC")
  expect_true(is.factor(typed_left$grp))
  expect_type(typed_left$payload, "list")
  expect_equal(typed_left$payload[[1]]$id, 1L)
  expect_true(is.na(typed_left$note[2]))

  typed_right <- fuzzystring_right_join(
    x_typed,
    y_typed,
    by = c(name = "approx_name"),
    max_dist = 1,
    distance_col = "distance"
  )

  expect_type(typed_right$meta, "list")
  expect_equal(typed_right$meta[[1]]$code, 10L)
  expect_true(is.na(typed_right$when[is.na(typed_right$name)][1]))

  # ------------------------------------------------------------------
  # data.frame inputs return plain data.frames
  # ------------------------------------------------------------------
  res_semi_df <- fuzzystring_semi_join(
    as.data.frame(a),
    as.data.frame(a),
    by = c(x = "x"),
    max_dist = 0
  )
  res_anti_df <- fuzzystring_anti_join(
    as.data.frame(a),
    as.data.frame(b),
    by = c(x = "y"),
    max_dist = 1
  )
  expect_s3_class(res_semi_df, "data.frame")
  expect_false(data.table::is.data.table(res_semi_df))
  expect_s3_class(res_anti_df, "data.frame")
  expect_false(data.table::is.data.table(res_anti_df))
})

test_that("tbl_df inputs return tbl_df outputs", {
  skip_if_not_installed("tibble")

  x_tbl <- tibble::tibble(
    name = c("Idea", "Premiom", "Very Good"),
    id = 1:3
  )

  y_tbl <- tibble::tibble(
    approx_name = c("Ideal", "Premium", "VeryGood"),
    grp = c("A", "B", "C")
  )

  res_inner <- fuzzystring_inner_join(
    x_tbl,
    y_tbl,
    by = c(name = "approx_name"),
    max_dist = 2,
    distance_col = "distance"
  )
  expect_s3_class(res_inner, "tbl_df")
  expect_false(data.table::is.data.table(res_inner))

  res_left <- fuzzystring_left_join(
    x_tbl,
    y_tbl,
    by = c(name = "approx_name"),
    max_dist = 2,
    distance_col = "distance"
  )
  expect_s3_class(res_left, "tbl_df")
  expect_false(data.table::is.data.table(res_left))

  res_semi <- fuzzystring_semi_join(
    x_tbl,
    y_tbl,
    by = c(name = "approx_name"),
    max_dist = 2
  )
  expect_s3_class(res_semi, "tbl_df")
  expect_false(data.table::is.data.table(res_semi))

  res_anti <- fuzzystring_anti_join(
    tibble::tibble(name = c("alpha", "bravo")),
    tibble::tibble(approx_name = c("charlie", "delta")),
    by = c(name = "approx_name"),
    max_dist = 1
  )
  expect_s3_class(res_anti, "tbl_df")
  expect_false(data.table::is.data.table(res_anti))
})

test_that("multi-column progressive join works correctly", {
  skip_if_not_installed("data.table")
  skip_if_not_installed("stringdist")

  a <- data.table::data.table(
    first = c("John", "Alice", "Bob"),
    last = c("Smith", "Smith", "Jones"),
    age = c(30, 25, 40)
  )
  b <- data.table::data.table(
    fname = c("Jon", "Alise", "Bob"),
    lname = c("Smit", "Smith", "Jones"),
    score = c(100, 200, 300)
  )

  # Inner join on first name and last name
  res <- fuzzystring_inner_join(
    a,
    b,
    by = c(first = "fname", last = "lname"),
    max_dist = 1,
    distance_col = "dist"
  )

  # John/Jon and Smith/Smit match (dist <= 1 on both)
  # Alice/Alise and Smith/Smith match (dist <= 1 on both)
  # Bob/Bob and Jones/Jones match (dist 0 on both)
  expect_equal(nrow(res), 3L)
  expect_equal(res$first, c("John", "Alice", "Bob"))
  expect_equal(res$fname, c("Jon", "Alise", "Bob"))

  # Semi join works
  res_semi <- fuzzystring_semi_join(
    a,
    b,
    by = c(first = "fname", last = "lname"),
    max_dist = 1
  )
  expect_equal(nrow(res_semi), 3L)
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

bind_by_rowid_cpp <- function(x_dt, y_dt, x_idx, y_idx, overlap) {
    .Call(`_fuzzystring_bind_by_rowid_cpp`, x_dt, y_dt, x_idx, y_idx, overlap)
}

bind_by_rowid_cpp_matches <- function(x_dt, y_dt, matches_dt, overlap) {
    .Call(`_fuzzystring_bind_by_rowid_cpp_matches`, x_dt, y_dt, matches_dt, overlap)
}

if (getRversion() >= "2.15.1") {
  utils::globalVariables(
    c(
      ".",
      ".I",
      ".N",
      ":=",
      "rowid",
      "key",
      "name",
      "newname",
      "..by2",
      "..extra_cols",
      "N",
      ".dummy",
      "as.formula",
      "_fuzzystring_bind_by_rowid_cpp",
      "_fuzzystring_bind_by_rowid_cpp_matches",
      "grp",
      "nchars",
      "value"
    )
  )
}

---
title: "Benchmarking fuzzystring against fuzzyjoin"
author: "Paul Efren Santos Andrade"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Benchmarking fuzzystring against fuzzyjoin}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 12,
  fig.height = 8
)
```

```{r setup}
library(dplyr)
library(tidyr)
library(ggplot2)

find_benchmark_csv <- function() {
  candidates <- c(
    file.path("inst", "extdata", "rbase_string_benchmark.csv"),
    file.path("..", "inst", "extdata", "rbase_string_benchmark.csv"),
    system.file("extdata", "rbase_string_benchmark.csv", package = "fuzzystring")
  )

  hit <- candidates[file.exists(candidates)][1]

  if (is.na(hit)) {
    stop(
      "Could not find the precomputed benchmark CSV. Expected it in ",
      paste(candidates, collapse = " or "),
      call. = FALSE
    )
  }

  hit
}

results_raw <- utils::read.csv(find_benchmark_csv(), stringsAsFactors = FALSE)

method_levels <- c(
  "osa", "lv", "dl", "hamming", "lcs",
  "qgram", "cosine", "jaccard", "jw", "soundex"
)

method_labels <- c(
  osa = "OSA",
  lv = "Levenshtein",
  dl = "Damerau-Levenshtein",
  hamming = "Hamming",
  lcs = "LCS",
  qgram = "Q-gram",
  cosine = "Cosine",
  jaccard = "Jaccard",
  jw = "Jaro-Winkler",
  soundex = "Soundex"
)

comparison_levels <- sprintf(
  "%.2f",
  sort(unique(results_raw$n_comps / 1e6))
)

summary_raw <- results_raw %>%
  group_by(expr, method, n_comps) %>%
  summarise(mean_time_ns = mean(time), .groups = "drop") %>%
  mutate(mean_time_ms = mean_time_ns / 1e6)

wide_summary <- summary_raw %>%
  select(expr, method, n_comps, mean_time_ms) %>%
  pivot_wider(names_from = expr, values_from = mean_time_ms) %>%
  mutate(
    runtime_ratio = fuzzyjoin / fuzzystring,
    method_label = factor(
      recode(method, !!!method_labels),
      levels = unname(method_labels[method_levels])
    ),
    comparisons_millions = n_comps / 1e6,
    comparisons_label = factor(
      sprintf("%.2f", comparisons_millions),
      levels = comparison_levels
    )
  ) %>%
  arrange(method_label, n_comps)

absolute_plot_data <- summary_raw %>%
  mutate(
    time_seconds = mean_time_ms / 1000,
    method_label = factor(
      recode(method, !!!method_labels),
      levels = unname(method_labels[method_levels])
    ),
    comparisons_label = factor(
      sprintf("%.2f", n_comps / 1e6),
      levels = comparison_levels
    ),
    implementation = factor(expr, levels = c("fuzzyjoin", "fuzzystring"))
  )

method_ranking <- wide_summary %>%
  group_by(method_label) %>%
  summarise(avg_runtime_ratio = mean(runtime_ratio), .groups = "drop") %>%
  arrange(desc(avg_runtime_ratio))

summary_table <- wide_summary %>%
  transmute(
    Method = as.character(method_label),
    `Candidate comparisons (M)` = sprintf("%.2f", comparisons_millions),
    `Mean time: fuzzyjoin (ms)` = round(fuzzyjoin, 2),
    `Mean time: fuzzystring (ms)` = round(fuzzystring, 2),
    `Runtime ratio (fuzzyjoin / fuzzystring)` = round(runtime_ratio, 2)
  )

ranking_table <- method_ranking %>%
  transmute(
    Method = as.character(method_label),
    `Average runtime ratio` = round(avg_runtime_ratio, 2)
  )

overall_ratio <- mean(wide_summary$runtime_ratio)
```

## Overview

This vignette documents the benchmark used to compare **fuzzyjoin** with the
`fuzzystring_join()` reimplementation in **fuzzystring**. The benchmark code is
shown for reproducibility, but it is **not executed** when this vignette is
built. Instead, the analysis below reads the precomputed CSV snapshot bundled in
`inst/extdata/rbase_string_benchmark.csv`.

The benchmark covers 10 distance methods, 3 sample sizes (`250`, `500`, and
`750`), and 10 repetitions per implementation. Across
`r nrow(wide_summary)` method-size combinations, **fuzzystring** is faster in
every case. The mean runtime ratio is `r sprintf("%.2f", overall_ratio)`x,
meaning that, on average, `fuzzyjoin` takes almost four times as long as the
reimplemented `fuzzystring` path on this benchmark snapshot.

## Benchmark Script

The script below is the benchmark used to generate the CSV analyzed in this
document. It is displayed as reference only.

```{r benchmark-script, eval = FALSE}
library(microbenchmark)
library(fuzzyjoin)
library(fuzzystring)
library(qdapDictionaries)
library(tibble)

samp_sizes <- c(250, 500, 750)
seed <- 2016

params <- list(
  list(method = "osa",     mode = "inner", max_dist = 1,   q = 0),
  list(method = "lv",      mode = "inner", max_dist = 1,   q = 0),
  list(method = "dl",      mode = "inner", max_dist = 1,   q = 0),
  list(method = "hamming", mode = "inner", max_dist = 1,   q = 0),
  list(method = "lcs",     mode = "inner", max_dist = 1,   q = 0),
  list(method = "qgram",   mode = "inner", max_dist = 2,   q = 2),
  list(method = "cosine",  mode = "inner", max_dist = 0.5, q = 2),
  list(method = "jaccard", mode = "inner", max_dist = 0.5, q = 2),
  list(method = "jw",      mode = "inner", max_dist = 0.5, q = 0),
  list(method = "soundex", mode = "inner", max_dist = 0.5, q = 0)
)

args <- commandArgs(trailingOnly = TRUE)
if (length(args) > 0) {
  params <- Filter(function(p) p$method %in% args, params)
}

data(misspellings)
words <- as.data.frame(DICTIONARY)

results <- data.frame()

for (p in params) {
  for (nsamp in samp_sizes) {
    cat(sprintf("Running %s with %d samples\n", p$method, nsamp))

    set.seed(seed)
    sub_misspellings <- misspellings[sample(seq_len(nrow(misspellings)), nsamp), ]

    bench <- microbenchmark(
      fuzzyjoin = {
        fuzzy_res <- stringdist_join(
          sub_misspellings, words,
          by         = c(misspelling = "word"),
          method     = p$method,
          mode       = p$mode,
          max_dist   = p$max_dist,
          q          = p$q
        )
      },
      fuzzystring = {
        fuzzystring_res <- fuzzystring_join(
          sub_misspellings, words,
          by       = c(misspelling = "word"),
          method   = p$method,
          mode     = p$mode,
          max_dist = p$max_dist,
          q        = p$q
        )
      },
      times = 10
    )

    # Verify that both implementations return the same result.
    fuzzy_res <- stringdist_join(
      sub_misspellings, words,
      by       = c(misspelling = "word"),
      method   = p$method,
      mode     = p$mode,
      max_dist = p$max_dist,
      q        = p$q
    )

    fuzzystring_res <- fuzzystring_join(
      sub_misspellings, words,
      by       = c(misspelling = "word"),
      method   = p$method,
      mode     = p$mode,
      max_dist = p$max_dist,
      q        = p$q
    )

    if (!isTRUE(all.equal(fuzzy_res, fuzzystring_res)) && p$method != "soundex") {
      message("Mismatch detected: fuzzyjoin vs fuzzystring for method: ", p$method)
    }

    df <- as.data.frame(bench)
    df$method <- p$method
    df$n_comps <- nrow(sub_misspellings) * nrow(words)
    df$os <- unname(Sys.info()["sysname"])

    results <- rbind(results, df)
  }
}

```

## Results

The benchmark spans 10 string distance methods and 3 sample sizes, producing 30
aggregated method-size combinations. The table below reports mean runtime in
milliseconds and the relative runtime ratio used throughout the figures.

```{r summary-table, include=FALSE}
knitr::kable(summary_table, align = "lrrrr")
```

The ranking below summarizes the average speedup by method. Larger values mean
that **fuzzystring** is faster by a wider margin.

```{r ranking-table, include=FALSE}
knitr::kable(ranking_table, align = "lr")
```

### Absolute Runtime

```{r absolute-runtime-plot}
ggplot(
  absolute_plot_data,
  aes(x = comparisons_label, y = time_seconds, fill = implementation)
) +
  geom_col(position = position_dodge(width = 0.75), width = 0.65) +
  facet_wrap(~ method_label, scales = "free_y") +
  labs(
    title = "Mean runtime by distance method",
    x = "Candidate comparisons (millions)",
    y = "Mean runtime (seconds)",
    fill = "Implementation"
  ) +
  scale_fill_manual(
    values = c("fuzzyjoin" = "#D95F02", "fuzzystring" = "#1B9E77"),
    labels = c("fuzzyjoin", "fuzzystring")
  ) +
  scale_y_continuous(labels = function(x) sprintf("%.1f", x)) +
  theme_minimal(base_size = 14) +
  theme(
    strip.text = element_text(face = "bold"),
    legend.position = "bottom",
    plot.title = element_text(face = "bold", hjust = 0.5),
    plot.subtitle = element_text(hjust = 0.5),
    axis.text.x = element_text(angle = 30, hjust = 1)
  )
```

### Relative Runtime

Values above 1 indicate that **fuzzystring** is faster than **fuzzyjoin** for
the same method and workload size.

```{r relative-runtime-plot}
ggplot(
  wide_summary,
  aes(x = comparisons_label, y = runtime_ratio)
) +
  geom_col(fill = "#D95F02", width = 0.55) +
  geom_hline(yintercept = 1, linetype = "dashed", color = "gray40") +
  facet_wrap(~ method_label, scales = "free_y") +
  labs(
    title = "Relative runtime against fuzzystring",
    subtitle = "Values above 1 mean fuzzystring is faster than fuzzyjoin",
    x = "Candidate comparisons (millions)",
    y = "Runtime ratio (fuzzyjoin / fuzzystring)"
  ) +
  scale_y_continuous(labels = function(x) sprintf("%.1f", x)) +
  theme_minimal(base_size = 14) +
  theme(
    strip.text = element_text(face = "bold"),
    legend.position = "none",
    plot.title = element_text(face = "bold", hjust = 0.5),
    plot.subtitle = element_text(hjust = 0.5),
    axis.text.x = element_text(angle = 30, hjust = 1)
  )
```

## Interpretation

This benchmark snapshot supports the same conclusion across all methods tested:
the **fuzzystring** reimplementation is consistently faster than **fuzzyjoin**
for the workloads covered here.

The largest average gains appear in Hamming, Levenshtein, OSA, and
Damerau-Levenshtein, while cosine, q-gram, jaccard, and soundex still show a
clear but smaller improvement. In other words, the reimplemented path does not
only match `fuzzyjoin` functionally; it also improves runtime across every
benchmarked method-size combination in this snapshot.

---
title: "Getting Started with fuzzystring"
author: "Paul Efren Santos Andrade"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with fuzzystring}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}

---


```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(fuzzystring)
```

## Introduction

**fuzzystring** provides fast, flexible fuzzy string joins for `data.frame` and
`data.table` objects using approximate string matching. Built on top of
`data.table` and `stringdist`, it uses compiled C++ result assembly plus
adaptive candidate planning to reduce unnecessary distance evaluations in
single-column joins.

## Installation

You can install **fuzzystring** from CRAN:

```r
install.packages("fuzzystring")
```

You can also install the development version from GitHub:

```r
# Using pak (recommended)
# pak::pak("PaulESantos/fuzzystring")

# Or using remotes
# remotes::install_github("PaulESantos/fuzzystring")
```

## Quick Start

Here's a simple example matching diamond cuts with slight misspellings:

```{r quick-start}
# Your messy data
x <- data.frame(
  name = c("Idea", "Premiom", "Very Good"), 
  id = 1:3
)

# Reference data
y <- data.frame(
  approx_name = c("Ideal", "Premium", "VeryGood"), 
  grp = c("A", "B", "C")
)

# Fuzzy join with max distance of 2 edits
fuzzystring_inner_join(
  x, y,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)
```

## Key Features

### All Join Types Supported

**fuzzystring** supports all standard join types. Below is a small, reusable
example dataset so you can compare the behavior of each join family.

```{r join-datasets}
x_join <- data.frame(
  name = c("Idea", "Premiom", "Very Good", "Gooood"),
  id = 1:4
)

y_join <- data.frame(
  approx_name = c("Ideal", "Premium", "VeryGood", "Good"),
  grp = c("A", "B", "C", "D")
)
```

- `fuzzystring_inner_join()`: Only matching rows.
- `fuzzystring_left_join()`: All rows from `x`, matching rows from `y`.
- `fuzzystring_right_join()`: All rows from `y`, matching rows from `x`.
- `fuzzystring_full_join()`: All rows from both tables.
- `fuzzystring_semi_join()`: Rows from `x` that have a match in `y`.
- `fuzzystring_anti_join()`: Rows from `x` that don't have a match in `y`.

#### Inner join

```{r join-inner, eval = TRUE}
fuzzystring_inner_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)
```

#### Left join

```{r join-left, eval = TRUE}
fuzzystring_left_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)
```

#### Right join

```{r join-right, eval = TRUE}
fuzzystring_right_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)
```

#### Full join

```{r join-full, eval = TRUE}
fuzzystring_full_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  distance_col = "distance"
)
```

#### Semi join (rows from `x` with a match in `y`)

```{r join-semi, eval = TRUE}
fuzzystring_semi_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2
)
```

#### Anti join (rows from `x` without a match in `y`)

```{r join-anti, eval = TRUE}
fuzzystring_anti_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2
)
```

#### Using the generic `fuzzystring_join()`

If you prefer a single entry point, you can use `fuzzystring_join()` directly
by specifying `mode`.

```{r join-generic, eval = TRUE}
fuzzystring_join(
  x_join, y_join,
  by = c(name = "approx_name"),
  max_dist = 2,
  mode = "left",
  distance_col = "distance"
)
```

### Multiple Distance Methods

You can choose from various distance metrics provided by the `stringdist` package:

```{r distance-methods, eval = FALSE}
# Optimal String Alignment (default)
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "osa")

# Damerau-Levenshtein
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "dl")

# Jaro-Winkler (good for names)
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "jw")

# Soundex (phonetic matching)
fuzzystring_inner_join(x, y, by = c(name = "approx_name"), method = "soundex")
```

### Case-Insensitive Matching

Use `ignore_case = TRUE` to ignore capitalization:

```{r ignore-case, eval = FALSE}
fuzzystring_inner_join(
  x, y, 
  by = c(name = "approx_name"),
  ignore_case = TRUE,
  max_dist = 1
)
```

## Advanced Usage

### Multiple Column Joins

You can match on multiple string columns at once. The same distance method and
threshold are applied to each mapped column.

```{r multi-column, eval = FALSE}
x_multi <- data.frame(
  first = c("Jon", "Maira"),
  last = c("Smyth", "Gonzales")
)

y_multi <- data.frame(
  first_ref = c("John", "Maria"),
  last_ref = c("Smith", "Gonzalez"),
  customer_id = 1:2
)

fuzzystring_inner_join(
  x_multi, y_multi,
  by = c(first = "first_ref", last = "last_ref"),
  method = "osa",
  max_dist = 1
)
```

## Performance

**fuzzystring** now keeps more of the join execution on a compiled C++ path
while using `data.table` to orchestrate candidate generation. In practice this
means compiled row expansion and binding across join modes, better preservation
of typed columns, and adaptive candidate planning that helps both
duplicate-heavy and low-duplication workloads.

For a dedicated comparison against `fuzzyjoin::stringdist_join()`, see the
benchmark article bundled with the package.

# distance column cannot corrupt input or internal columns

    Code
      fuzzystring_inner_join(x, y, by = c(value = "approx"), distance_col = "x")
    Condition
      Error:
      ! `distance_col` must not collide with input or internal columns.

---

    Code
      fuzzystring_inner_join(x, y, by = c(value = "approx"), distance_col = "value")
    Condition
      Error:
      ! `distance_col` must not collide with input or internal columns.

# public arguments are validated before join planning

    Code
      fuzzystring_join(x, y, by = c(value = "approx"), max_dist = NA_real_)
    Condition
      Error:
      ! `max_dist` must be a single finite non-negative number.

---

    Code
      fuzzystring_join(x, y, by = c(value = "approx"), max_dist = Inf)
    Condition
      Error:
      ! `max_dist` must be a single finite non-negative number.

---

    Code
      fuzzystring_join(x, y, by = c(value = "approx"), ignore_case = "yes")
    Condition
      Error:
      ! `ignore_case` must be `TRUE` or `FALSE`.

---

    Code
      fuzzystring_join(x, y, by = character())
    Condition
      Error:
      ! `by` must specify at least one valid pair of columns.

---

    Code
      fuzzystring_join(x, y, by = c(missing = "approx"))
    Condition
      Error:
      ! Join columns not found: x$missing.

# Cartesian expansion rejects unsupported result sizes

    Code
      fst_checked_cartesian_sizes(.Machine$integer.max, 2L)
    Condition
      Error:
      ! The join result is too large to represent safely.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dt_stringdist_join.R
\name{fuzzystring_join}
\alias{fuzzystring_join}
\alias{fuzzystring_inner_join}
\alias{fuzzystring_left_join}
\alias{fuzzystring_right_join}
\alias{fuzzystring_full_join}
\alias{fuzzystring_semi_join}
\alias{fuzzystring_anti_join}
\title{Join two tables based on fuzzy string matching}
\usage{
fuzzystring_join(
  x,
  y,
  by = NULL,
  max_dist = 2,
  method = c("osa", "lv", "dl", "hamming", "lcs", "qgram", "cosine", "jaccard", "jw",
    "soundex"),
  mode = "inner",
  ignore_case = FALSE,
  distance_col = NULL,
  ...
)

fuzzystring_inner_join(x, y, by = NULL, distance_col = NULL, ...)

fuzzystring_left_join(x, y, by = NULL, distance_col = NULL, ...)

fuzzystring_right_join(x, y, by = NULL, distance_col = NULL, ...)

fuzzystring_full_join(x, y, by = NULL, distance_col = NULL, ...)

fuzzystring_semi_join(x, y, by = NULL, distance_col = NULL, ...)

fuzzystring_anti_join(x, y, by = NULL, distance_col = NULL, ...)
}
\arguments{
\item{x}{A \code{data.frame} or \code{data.table}.}

\item{y}{A \code{data.frame} or \code{data.table}.}

\item{by}{Columns by which to join the two tables. You can supply a character
vector of common names (e.g. \code{c("name")} ), or a named vector mapping
\code{x} to \code{y} (e.g. \code{c(name = "approx_name")}).}

\item{max_dist}{Maximum distance to use for joining. Smaller values are stricter.}

\item{method}{Method for computing string distance, see
\code{?stringdist::stringdist} and the \code{stringdist} package vignettes.}

\item{mode}{One of \code{"inner"}, \code{"left"}, \code{"right"}, \code{"full"},
\code{"semi"}, or \code{"anti"}.}

\item{ignore_case}{Logical; if \code{TRUE}, comparisons are case-insensitive.}

\item{distance_col}{If not \code{NULL}, adds a column with this name containing
the computed distance for each matched pair (or \code{NA} for unmatched rows
in outer joins).}

\item{...}{Additional arguments passed to \code{\link[stringdist]{stringdist}}.}
}
\value{
A joined table that preserves the container class of \code{x}:
\code{data.table} inputs return \code{data.table}, tibble inputs return
tibble, and plain \code{data.frame} inputs return plain
\code{data.frame}. See \code{\link{fuzzystring_join_backend}} for details
on output structure.
}
\description{
Uses \code{stringdist::stringdist()} to compute distances and a
\code{data.table}-orchestrated backend with compiled 'C++' assembly to produce
the final result. This is the main user-facing entry point for fuzzy joins on
strings.
}
\details{
If \code{method = "soundex"}, \code{max_dist} is automatically set to 0.5,
since Soundex distance is 0 (match) or 1 (no match).

When \code{by} maps multiple columns, the same \code{method},
\code{max_dist}, and any additional \code{stringdist} arguments are applied
independently to each mapped column, and a row pair is kept only when all
mapped columns satisfy the distance threshold.
When \code{distance_col} is requested for a multi-column join, it contains the
maximum distance across the mapped columns for each matched row pair.

For single-column joins, fuzzystring uses adaptive candidate planning before
calling \code{stringdist::stringdist()}. For Levenshtein-like methods
(\code{"osa"}, \code{"lv"}, \code{"dl"}), a fast prefilter is applied: if
\code{abs(nchar(v1) - nchar(v2)) > max_dist}, the pair cannot match, so
distance is not computed for that pair. For low-duplication workloads, the
planner can also evaluate larger dense blocks of unique values to reduce
orchestration overhead while preserving the same matching semantics.
When \code{max_dist = 0} and no additional metric arguments are supplied,
methods whose zero distance implies string equality use an exact keyed join
instead of evaluating the full candidate distance set.
}
\examples{
\donttest{
if (requireNamespace("ggplot2", quietly = TRUE)) {
  d <- data.table::data.table(approximate_name = c("Idea", "Premiom"))
  # Match diamonds$cut to d$approximate_name
  res <- fuzzystring_inner_join(ggplot2::diamonds, d,
    by = c(cut = "approximate_name"),
    max_dist = 1
  )
  head(res)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fuzzystring-package.R
\docType{package}
\name{fuzzystring}
\alias{fuzzystring-package}
\alias{fuzzystring}
\title{fuzzystring: Fast fuzzy string joins for data frames}
\description{
\code{fuzzystring} provides fuzzy inner, left, right, full, semi, and anti joins
for \code{data.frame} and \code{data.table} objects using approximate string matching.
It combines \code{stringdist} metrics with a \code{data.table} backend and compiled 'C++'
result assembly to reduce overhead in large joins while preserving familiar
join semantics.
}
\details{
Main entry points are \code{\link[=fuzzystring_join]{fuzzystring_join()}} and the convenience wrappers
\code{\link[=fuzzystring_inner_join]{fuzzystring_inner_join()}}, \code{\link[=fuzzystring_left_join]{fuzzystring_left_join()}},
\code{\link[=fuzzystring_right_join]{fuzzystring_right_join()}}, \code{\link[=fuzzystring_full_join]{fuzzystring_full_join()}},
\code{\link[=fuzzystring_semi_join]{fuzzystring_semi_join()}}, and \code{\link[=fuzzystring_anti_join]{fuzzystring_anti_join()}}.

The package also includes the example dataset \link{misspellings}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/PaulESantos/fuzzystring}
  \item \url{https://paulesantos.github.io/fuzzystring/}
  \item Report bugs at \url{https://github.com/PaulESantos/fuzzystring/issues}
}

}
\author{
\strong{Maintainer}: Paul E. Santos Andrade \email{paulefrens@gmail.com} (\href{https://orcid.org/0000-0002-6635-0375}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Paul E. Santos Andrade \email{paulefrens@gmail.com} (\href{https://orcid.org/0000-0002-6635-0375}{ORCID}) [copyright holder]
}

Other contributors:
\itemize{
  \item David Robinson \email{admiral.david@gmail.com} (aut of fuzzyjoin) [contributor]
}

}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misspellings.R
\docType{data}
\name{misspellings}
\alias{misspellings}
\title{A corpus of common misspellings, for examples and practice}
\format{
An object of class \code{tbl_df} (inherits from \code{tbl}, \code{data.frame}) with 4505 rows and 2 columns.
}
\source{
\url{https://en.wikipedia.org/wiki/Wikipedia:Lists_of_common_misspellings/For_machines}
}
\usage{
misspellings
}
\description{
This is a \code{tbl_df} mapping misspellings of their words, compiled by
Wikipedia, where it is licensed under the CC-BY SA license. (Three words with
non-ASCII characters were filtered out). If you'd like to reproduce this
dataset from Wikipedia, see the example code below.
}
\examples{

\donttest{
if (interactive()) {
library(rvest)
library(readr)
library(dplyr)
library(stringr)
library(tidyr)

u <- "https://en.wikipedia.org/wiki/Wikipedia:Lists_of_common_misspellings/For_machines"
h <- read_html(u)

misspellings <- h \%>\%
  html_nodes("pre") \%>\%
  html_text() \%>\%
  read_delim(col_names = c("misspelling", "correct"),
                    delim = ">",
                    skip = 1) \%>\%
  mutate(misspelling = str_sub(misspelling,
                                               1, -2)) |>
  separate_rows(correct, sep = ", ") |>
  filter(Encoding(correct) != "UTF-8")
}
}

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dt_fuzzy_join_cpp.R
\name{fuzzystring_join_backend}
\alias{fuzzystring_join_backend}
\title{Fuzzy join backend using 'data.table' plus compiled 'C++' assembly}
\usage{
fuzzystring_join_backend(
  x,
  y,
  by = NULL,
  match_fun = NULL,
  multi_by = NULL,
  multi_match_fun = NULL,
  index_match_fun = NULL,
  mode = "inner",
  ...
)
}
\arguments{
\item{x}{A \code{data.frame} or \code{data.table}.}

\item{y}{A \code{data.frame} or \code{data.table}.}

\item{by}{Columns by which to join the two tables. See
\code{\link{fuzzystring_join}}.}

\item{match_fun}{A function used to match values. It must return a logical
vector (or a data.frame/data.table whose first column is logical) indicating
which pairs match. For multi-column joins, you may pass a list of functions
(one per column).}

\item{multi_by}{A character vector of column names used for multi-column
matching when \code{multi_match_fun} is supplied.}

\item{multi_match_fun}{A function that receives matrices of unique values for
\code{x} and \code{y} (rows correspond to unique combinations of \code{multi_by}).
It must return a logical vector (or a data.frame/data.table whose first column
is logical) indicating which rows match.}

\item{index_match_fun}{A function that receives the joined columns from
\code{x} and \code{y} and returns a table with integer columns \code{x} and
\code{y} (1-based row indices).}

\item{mode}{One of \code{"inner"}, \code{"left"}, \code{"right"}, \code{"full"},
\code{"semi"}, or \code{"anti"}.}

\item{...}{Additional arguments passed to the matching function(s).}
}
\value{
A joined table that preserves the container class of \code{x}:
\code{data.table} inputs return \code{data.table}, tibble inputs return
tibble, and plain \code{data.frame} inputs return plain
\code{data.frame}. See \code{\link{fuzzystring_join}}.
}
\description{
Low-level engine used by \code{\link{fuzzystring_join}} and the compiled
fuzzy join helpers. It builds the match index with R 'data.table' and then
expands and assembles the result using compiled 'C++' binding code for speed.
}
\details{
This function works like \code{\link{fuzzystring_join}}, but replaces the
R-based output assembly with a compiled 'C++' implementation. This provides
better performance, especially for large joins, outer joins, and wide tables.
It is intended as a backend and does not compute distances itself; use
\code{\link{fuzzystring_join}} for string-distance based matching.

The C++ implementation handles:
\itemize{
\item Efficient subsetting by row indices
\item Proper handling of NA values in outer joins
\item Type-safe column operations for all common R types
\item Preservation of factor levels and attributes
\item Column name conflicts with .x/.y suffixes
}
}
\keyword{internal}
